#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-11-01
# @Filename: modelcube.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-14 11:48:16


from __future__ import absolute_import, division, print_function

import distutils
import warnings

import numpy as np
from pkg_resources import parse_version
from astropy.io import fits
from astropy.wcs import WCS

import marvin
import marvin.core.exceptions
import marvin.tools.maps
import marvin.tools.spaxel
import marvin.utils.general.general
from marvin.core.exceptions import MarvinError
from marvin.tools.quantities import DataCube, Map, Spectrum
from marvin.utils.datamodel.dap import Model, datamodel
from marvin.utils.general import FuzzyDict, gunzip, check_versions

from .core import MarvinToolsClass
from .mixins import DAPallMixIn, GetApertureMixIn, NSAMixIn


class ModelCube(MarvinToolsClass, NSAMixIn, DAPallMixIn, GetApertureMixIn):
    """A class to interface with MaNGA DAP model cubes.

    This class represents a DAP model cube, initialised either from a file,
    a database, or remotely via the Marvin API. In addition to
    the parameters and variables defined for `~.MarvinToolsClass`, the
    following parameters and attributes are specific to `.Maps`.

    Parameters:
        bintype (str or None):
            The binning type. For MPL-4, one of the following: ``'NONE',
            'RADIAL', 'STON'`` (if ``None`` defaults to ``'NONE'``).
            For MPL-5, one of, ``'ALL', 'NRE', 'SPX', 'VOR10'``
            (defaults to ``'SPX'``). MPL-6 also accepts the ``'HYB10'`` binning
            schema.
        template (str or None):
            The stellar template used. For MPL-4, one of
            ``'M11-STELIB-ZSOL', 'MILES-THIN', 'MIUSCAT-THIN'`` (if ``None``,
            defaults to ``'MIUSCAT-THIN'``). For MPL-5 and successive, the only
            option in ``'GAU-MILESHC'`` (``None`` defaults to it).

    Attributes:
        header (`astropy.io.fits.Header`):
            The header of the datacube.
        wcs (`astropy.wcs.WCS`):
            The WCS solution for this plate

    """

    def __init__(self, input=None, filename=None, mangaid=None, plateifu=None,
                 mode=None, data=None, release=None,
                 drpall=None, download=None, nsa_source='auto',
                 bintype=None, template=None, template_kin=None):

        if template_kin is not None:
            warnings.warn('template_kin is deprecated and will be removed in a future version.',
                          DeprecationWarning)
            template = template_kin if template is None else template

        # _set_datamodel will replace these strings with datamodel objects.
        self.bintype = bintype
        self.template = template
        self.datamodel = None

        self._bitmasks = None

        MarvinToolsClass.__init__(self, input=input, filename=filename,
                                  mangaid=mangaid, plateifu=plateifu,
                                  mode=mode, data=data, release=release,
                                  drpall=drpall, download=download)

        NSAMixIn.__init__(self, nsa_source=nsa_source)

        # Checks that DAP is at least MPL-5
        MPL5 = distutils.version.StrictVersion('2.0.2')
        if self.filename is None and distutils.version.StrictVersion(self._dapver) < MPL5:
            raise MarvinError('ModelCube requires at least dapver=\'2.0.2\'')

        self.header = None
        self.wcs = None
        self._wavelength = None
        self._redcorr = None
        self._shape = None

        # Model extensions
        self._extension_data = {}
        self._binned_flux = None
        self._redcorr = None
        self._full_fit = None
        self._emline_fit = None
        self._stellarcont_fit = None

        if self.data_origin == 'file':
            self._load_modelcube_from_file()
        elif self.data_origin == 'db':
            self._load_modelcube_from_db()
        elif self.data_origin == 'api':
            self._load_modelcube_from_api()
        else:
            raise marvin.core.exceptions.MarvinError(
                'data_origin={0} is not valid'.format(self.data_origin))

        # Confirm that drpver and dapver match the ones from the header.
        marvin.tools.maps.Maps._check_versions(self)

    def __repr__(self):
        """Representation for ModelCube."""

        return ('<Marvin ModelCube (plateifu={0!r}, mode={1!r}, data_origin={2!r}, '
                'bintype={3!r}, template={4!r})>'.format(self.plateifu,
                                                         self.mode,
                                                         self.data_origin,
                                                         str(self.bintype),
                                                         str(self.template)))

    def __getitem__(self, xy):
        """Returns the spaxel for ``(x, y)``"""

        return self.getSpaxel(x=xy[1], y=xy[0], xyorig='lower')

    def _set_datamodel(self):
        """Sets the datamodel, template, and bintype."""

        self.datamodel = datamodel[self.release].models
        self._bitmasks = datamodel[self.release].bitmasks
        self.bintype = self.datamodel.parent.get_bintype(self.bintype)
        self.template = self.datamodel.parent.get_template(self.template)

    def _getFullPath(self):
        """Returns the full path of the file in the tree."""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')
        daptype = '{0}-{1}'.format(self.bintype, self.template)

        return super(ModelCube, self)._getFullPath('mangadap', ifu=ifu,
                                                   drpver=self._drpver,
                                                   dapver=self._dapver,
                                                   plate=plate, mode='LOGCUBE',
                                                   daptype=daptype)

    def download(self):
        """Downloads the cube using sdss_access - Rsync"""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')
        daptype = '{0}-{1}'.format(self.bintype, self.template)

        return super(ModelCube, self).download('mangadap', ifu=ifu,
                                               drpver=self._drpver,
                                               dapver=self._dapver,
                                               plate=plate, mode='LOGCUBE',
                                               daptype=daptype)

    def _load_modelcube_from_file(self):
        """Initialises a model cube from a file."""

        if self.data is not None:
            assert isinstance(self.data, fits.HDUList), 'data is not an HDUList object'
        else:
            try:
                with gunzip(self.filename) as gg:
                    self.data = fits.open(gg.name)
            except IOError as err:
                raise IOError('filename {0} cannot be found: {1}'.format(self.filename, err))

        self.header = self.data[0].header
        self._check_file(self.header, self.data, 'ModelCube')
        self.wcs = WCS(self.data['FLUX'].header)
        self._wavelength = self.data['WAVE'].data
        self._redcorr = self.data['REDCORR'].data
        self._shape = (self.data['FLUX'].header['NAXIS2'],
                       self.data['FLUX'].header['NAXIS1'])

        self.plateifu = self.header['PLATEIFU']
        self.mangaid = self.header['MANGAID']

        # Checks and populates release.
        file_drpver = self.header['VERSDRP3']
        file_drpver = 'v1_5_1' if file_drpver == 'v1_5_0' else file_drpver

        file_ver = marvin.config.lookUpRelease(file_drpver)
        assert file_ver is not None, 'cannot find file version.'

        if file_drpver != self._drpver:
            warnings.warn('mismatch between file version={0} and object release={1}. '
                          'Setting object release to {0}'.format(file_ver, self._release),
                          marvin.core.exceptions.MarvinUserWarning)
            self._release = file_ver

        self._drpver, self._dapver = marvin.config.lookUpVersions(release=self._release)

        # Updates datamodel, bintype, and template with the versions from the header.
        self.datamodel = datamodel[self._dapver].models
        self.bintype = self.datamodel.parent.get_bintype(self.header['BINKEY'].strip().upper())
        if check_versions(self._dapver, datamodel['MPL-8'].release):
            tempkey = self.header['DAPTYPE'].split('-', 1)[-1]
        else:
            tempkey = self.header['SCKEY']
        self.template = self.datamodel.parent.get_template(tempkey.strip().upper())

    def _load_modelcube_from_db(self):
        """Initialises a model cube from the DB."""

        mdb = marvin.marvindb
        plate, ifu = self.plateifu.split('-')

        if not mdb.isdbconnected:
            raise MarvinError('No db connected')

        else:

            datadb = mdb.datadb
            dapdb = mdb.dapdb

            dm = datamodel[self.release]
            if dm.db_only:
                if self.bintype not in dm.db_only:
                    raise marvin.core.exceptions.MarvinError(
                        'Specified bintype {0} is not '
                        'available in the DB'.format(self.bintype.name))

            if self.data:
                assert isinstance(self.data, dapdb.ModelCube), \
                    'data is not an instance of marvindb.dapdb.ModelCube.'
            else:
                # Initial query for version
                version_query = mdb.session.query(dapdb.ModelCube).join(
                    dapdb.File,
                    datadb.PipelineInfo,
                    datadb.PipelineVersion).filter(
                        datadb.PipelineVersion.version == self._dapver).from_self()

                # Query for model cube parameters
                db_modelcube = version_query.join(
                    dapdb.File,
                    datadb.Cube,
                    datadb.IFUDesign).filter(
                        datadb.Cube.plate == plate,
                        datadb.IFUDesign.name == str(ifu)).from_self().join(
                            dapdb.File,
                            dapdb.FileType).filter(dapdb.FileType.value == 'LOGCUBE').join(
                                dapdb.Structure, dapdb.BinType).join(
                                    dapdb.Template,
                                    dapdb.Structure.template_kin_pk == dapdb.Template.pk).filter(
                                        dapdb.BinType.name == self.bintype.name,
                                        dapdb.Template.name == self.template.name).use_cache(self.cache_region).all()

                if len(db_modelcube) > 1:
                    raise MarvinError('more than one ModelCube found for '
                                      'this combination of parameters.')

                elif len(db_modelcube) == 0:
                    raise MarvinError('no ModelCube found for this combination of parameters.')

                self.data = db_modelcube[0]

            self.header = self.data.file.primary_header
            self.wcs = WCS(self.data.file.cube.wcs.makeHeader())
            self._wavelength = np.array(self.data.file.cube.wavelength.wavelength, dtype=np.float)
            self._redcorr = np.array(self.data.redcorr[0].value, dtype=np.float)
            self._shape = self.data.file.cube.shape.shape

            self.plateifu = str(self.header['PLATEIFU'].strip())
            self.mangaid = str(self.header['MANGAID'].strip())

    def _load_modelcube_from_api(self):
        """Initialises a model cube from the API."""

        url = marvin.config.urlmap['api']['getModelCube']['url']
        url_full = url.format(name=self.plateifu, bintype=self.bintype.name,
                              template=self.template.name)

        try:
            response = self._toolInteraction(url_full)
        except Exception as ee:
            raise MarvinError('found a problem when checking if remote model cube '
                              'exists: {0}'.format(str(ee)))

        data = response.getData()

        self.header = fits.Header.fromstring(data['header'])
        self.wcs = WCS(fits.Header.fromstring(data['wcs_header']))
        self._wavelength = np.array(data['wavelength'])
        self._redcorr = np.array(data['redcorr'])
        self._shape = tuple(data['shape'])

        self.plateifu = str(self.header['PLATEIFU'].strip())
        self.mangaid = str(self.header['MANGAID'].strip())

    def getSpaxel(self, x=None, y=None, ra=None, dec=None,
                  cube=False, maps=False, **kwargs):
        """Returns the :class:`~marvin.tools.spaxel.Spaxel` matching certain coordinates.

        The coordinates of the spaxel to return can be input as ``x, y`` pixels
        relative to``xyorig`` in the cube, or as ``ra, dec`` celestial
        coordinates.

        If ``spectrum=True``, the returned |spaxel| will be instantiated with the
        DRP spectrum of the spaxel for the DRP cube associated with this
        ModelCube. The same is true for ``properties=True`` for the DAP
        properties of the spaxel in the Maps associated with these coordinates.

        Parameters:
            x,y (int or array):
                The spaxel coordinates relative to ``xyorig``. If ``x`` is an
                array of coordinates, the size of ``x`` must much that of
                ``y``.
            ra,dec (float or array):
                The coordinates of the spaxel to return. The closest spaxel to
                those coordinates will be returned. If ``ra`` is an array of
                coordinates, the size of ``ra`` must much that of ``dec``.
            xyorig ({'center', 'lower'}):
                The reference point from which ``x`` and ``y`` are measured.
                Valid values are ``'center'`` (default), for the centre of the
                spatial dimensions of the cube, or ``'lower'`` for the
                lower-left corner. This keyword is ignored if ``ra`` and
                ``dec`` are defined.
            cube (bool):
                If ``True``, the |spaxel| will be initialised with the
                corresponding DRP cube data.
            maps (bool):
                If ``True``, the |spaxel| will be initialised with the
                corresponding DAP Maps properties for this spaxel.

        Returns:
            spaxels (list):
                The |spaxel|_ objects for this cube/maps corresponding to the
                input coordinates. The length of the list is equal to the
                number of input coordinates.

        .. |spaxel| replace:: :class:`~marvin.tools.spaxel.Spaxel`

        """

        for old_param in ['drp', 'properties']:
            if old_param in kwargs:
                raise marvin.core.exceptions.MarvinDeprecationError(
                    'the {0} parameter has been deprecated. '
                    'Use cube or maps.'.format(old_param))

        return marvin.utils.general.general.getSpaxel(
            x=x, y=y, ra=ra, dec=dec,
            cube=cube, maps=maps, modelcube=self, **kwargs)

    def _get_extension_data(self, name, ext=None):
        """Returns the data from an extension."""

        model = self.datamodel[name]
        ext_name = model.fits_extension(ext)

        if ext_name in self._extension_data:
            return self._extension_data[ext_name]

        if self.data_origin == 'file':
            ext_data = self.data[model.fits_extension(ext)].data

        elif self.data_origin == 'db':
            # If the table is "spaxel", this must be a 3D cube. If it is "cube",
            # uses self.data, which is basically the DataModelClass.Cube instance.
            ext_data = self.data.get3DCube(model.db_column(ext))

        elif self.data_origin == 'api':

            params = {'release': self._release}
            url = marvin.config.urlmap['api']['getModelCubeExtension']['url']

            try:
                response = self._toolInteraction(
                    url.format(name=self.plateifu,
                               modelcube_extension=model.fits_extension(ext).lower(),
                               bintype=self.bintype.name, template=self.template.name),
                    params=params)
            except Exception as ee:
                raise MarvinError('found a problem when checking if remote '
                                  'modelcube exists: {0}'.format(str(ee)))

            data = response.getData()
            cube_ext_data = data['extension_data']
            ext_data = np.array(cube_ext_data) if cube_ext_data is not None else None

        self._extension_data[ext_name] = ext_data

        return ext_data

    def _get_spaxel_quantities(self, x, y, spaxel=None):
        """Returns a dictionary of spaxel quantities."""

        modelcube_quantities = FuzzyDict({})

        if self.data_origin == 'db':

            session = marvin.marvindb.session
            dapdb = marvin.marvindb.dapdb

        if self.data_origin == 'file' or self.data_origin == 'db':

            _db_row = None

            for dm in self.datamodel:

                data = {'value': None, 'ivar': None, 'mask': None}

                for key in data:

                    if key == 'ivar' and not dm.has_ivar():
                        continue
                    if key == 'mask' and not dm.has_mask():
                        continue

                    if self.data_origin == 'file':

                        extname = dm.fits_extension(None if key == 'value' else key)
                        data[key] = self.data[extname].data[:, y, x]

                    elif self.data_origin == 'db':

                        colname = dm.db_column(None if key == 'value' else key)

                        if not _db_row:
                            _db_row = session.query(dapdb.ModelSpaxel).filter(
                                dapdb.ModelSpaxel.modelcube_pk == self.data.pk,
                                dapdb.ModelSpaxel.x == x, dapdb.ModelSpaxel.y == y).use_cache(self.cache_region).one()

                        data[key] = np.array(getattr(_db_row, colname))

                quantity = Spectrum(data['value'], ivar=data['ivar'], mask=data['mask'],
                                    wavelength=self._wavelength, unit=dm.unit,
                                    pixmask_flag=dm.pixmask_flag)

                if spaxel:
                    quantity._init_bin(spaxel=spaxel, parent=self, datamodel=dm)

                modelcube_quantities[dm.full()] = quantity

        if self.data_origin == 'api':

            params = {'release': self._release}
            url = marvin.config.urlmap['api']['getModelCubeQuantitiesSpaxel']['url']

            try:
                response = self._toolInteraction(url.format(name=self.plateifu,
                                                            x=x, y=y,
                                                            bintype=self.bintype.name,
                                                            template=self.template.name,
                                                            params=params))
            except Exception as ee:
                raise MarvinError('found a problem when checking if remote modelcube '
                                  'exists: {0}'.format(str(ee)))

            data = response.getData()

            for dm in self.datamodel:

                quantity = Spectrum(data[dm.name]['value'], ivar=data[dm.name]['ivar'],
                                    mask=data[dm.name]['mask'], wavelength=data['wavelength'],
                                    unit=dm.unit, pixmask_flag=dm.pixmask_flag)

                if spaxel:
                    quantity._init_bin(spaxel=spaxel, parent=self, datamodel=dm)

                modelcube_quantities[dm.full()] = quantity

        return modelcube_quantities

    def get_binid(self, model=None):
        """Returns the binid map associated with a model.

        Parameters
        ----------
        model : `datamodel.Model` or None
            The model for which the associated binid map will be returned.
            If ``binid=None``, the default binid is returned.

        Returns
        -------
        binid : `Map`
            A `Map` with the binid associated with ``model`` or the default
            binid.

        """

        assert model is None or isinstance(model, Model), 'invalid model type.'

        if model is not None:
            binid_prop = model.binid
        else:
            binid_prop = self.datamodel.parent.default_binid

        # Before MPL-6, the modelcube does not include the binid extension,
        # so we need to get the binid map from the associated MAPS.
        if (distutils.version.StrictVersion(self._dapver) <
                distutils.version.StrictVersion('2.1')):
            return self.getMaps().get_binid()

        if self.data_origin == 'file':

            if binid_prop.channel is None:
                binid_map_data = self.data[binid_prop.name].data[:, :]
            else:
                binid_map_data = self.data[binid_prop.name].data[binid_prop.channel.idx, :, :]

        elif self.data_origin == 'db':

            mdb = marvin.marvindb

            table = mdb.dapdb.ModelSpaxel
            column = getattr(table, binid_prop.db_column())

            binid_list = mdb.session.query(column).filter(
                table.modelcube_pk == self.data.pk).order_by(table.x, table.y).all()

            nx = ny = int(np.sqrt(len(binid_list)))
            binid_array = np.array(binid_list)

            binid_map_data = binid_array.transpose().reshape((ny, nx)).transpose(1, 0)

        elif self.data_origin == 'api':

            params = {'release': self._release}
            url = marvin.config.urlmap['api']['getModelCubeBinid']['url']

            extension = model.fits_extension().lower() if model is not None else 'flux'

            try:
                response = self._toolInteraction(
                    url.format(name=self.plateifu,
                               modelcube_extension=extension,
                               bintype=self.bintype.name,
                               template=self.template.name), params=params)
            except Exception as ee:
                raise MarvinError('found a problem when checking if remote '
                                  'modelcube exists: {0}'.format(str(ee)))

            if response.results['error'] is not None:
                raise MarvinError('found a problem while getting the binid from API: {}'
                                  .format(str(response.results['error'])))

            binid_map_data = np.array(response.getData()['binid'])

        binid_map = Map(binid_map_data, unit=binid_prop.unit)
        binid_map._datamodel = binid_prop

        return binid_map

    @property
    def binned_flux(self):
        """Returns the binned flux datacube."""

        model = self.datamodel['binned_flux']

        binned_flux_array = self._get_extension_data('flux')
        binned_flux_ivar = self._get_extension_data('flux', 'ivar')
        binned_flux_mask = self._get_extension_data('flux', 'mask')

        return DataCube(binned_flux_array,
                        np.array(self._wavelength),
                        ivar=binned_flux_ivar,
                        mask=binned_flux_mask,
                        redcorr=self._redcorr,
                        binid=self.get_binid(model),
                        unit=model.unit,
                        pixmask_flag=model.pixmask_flag)

    @property
    def full_fit(self):
        """Returns the full fit datacube."""

        model = self.datamodel['full_fit']

        model_array = self._get_extension_data('full_fit')
        model_mask = self._get_extension_data('flux', 'mask')

        return DataCube(model_array,
                        np.array(self._wavelength),
                        ivar=None,
                        mask=model_mask,
                        redcorr=self._redcorr,
                        binid=self.get_binid(model),
                        unit=model.unit,
                        pixmask_flag=model.pixmask_flag)

    @property
    def emline_fit(self):
        """Returns the emission line fit."""

        model = self.datamodel['emline_fit']

        emline_array = self._get_extension_data('emline_fit')
        if model.has_mask():
            emline_mask = self._get_extension_data('emline_fit', 'mask')
        else:
            emline_mask = None

        return DataCube(emline_array,
                        np.array(self._wavelength),
                        ivar=None,
                        mask=emline_mask,
                        redcorr=self._redcorr,
                        binid=self.get_binid(model),
                        unit=model.unit,
                        pixmask_flag=model.pixmask_flag)

    @property
    def stellarcont_fit(self):
        """Returns the stellar continuum fit."""

        isMPL8 = check_versions(self._dapver, datamodel['MPL-8'].release)

        if isMPL8:
            array = self._get_extension_data('stellar_fit')
            model = self.datamodel['stellar_fit']
            mask = self._get_extension_data('stellar_fit', 'mask')
        else:
            array = (self._get_extension_data('full_fit') -
                     self._get_extension_data('emline_fit') -
                     self._get_extension_data('emline_base_fit'))
            model = self.datamodel['full_fit']
            mask = self._get_extension_data('flux', 'mask')

        return DataCube(array,
                        np.array(self._wavelength),
                        ivar=None,
                        mask=mask,
                        redcorr=self._redcorr,
                        binid=self.get_binid(model),
                        unit=model.unit,
                        pixmask_flag=model.pixmask_flag)

    @property
    def lsf(self):
        """Returns the pre-pixelized LSF"""

        isMPL10 = check_versions(self._dapver, datamodel['MPL-10'].release)
        if not isMPL10:
            raise MarvinError('LSF property only available for MPLs >= 10')

        model = self.datamodel['lsf']

        lsf_array = self._get_extension_data('lsf')
        if model.has_mask():
            lsf_mask = self._get_extension_data('lsf', 'mask')
        else:
            lsf_mask = None

        return DataCube(lsf_array,
                        np.array(self._wavelength),
                        ivar=None,
                        mask=lsf_mask,
                        redcorr=self._redcorr,
                        binid=self.get_binid(model),
                        unit=model.unit,
                        pixmask_flag=model.pixmask_flag)

    def getCube(self):
        """Returns the associated `~marvin.tools.cube.Cube`."""

        if self.data_origin == 'db':
            cube_data = self.data.file.cube
        else:
            cube_data = None

        return marvin.tools.cube.Cube(data=cube_data,
                                      plateifu=self.plateifu,
                                      release=self.release)

    def getMaps(self):
        """Returns the associated`~marvin.tools.maps.Maps`."""

        return marvin.tools.maps.Maps(plateifu=self.plateifu,
                                      bintype=self.bintype,
                                      template=self.template,
                                      release=self.release)

    def is_binned(self):
        """Returns True if the ModelCube is not unbinned."""

        return self.bintype.binned

    def get_unbinned(self):
        """Returns a version of ``self`` corresponding to the unbinned ModelCube."""

        if not self.is_binned:
            return self
        else:
            unbinned = self.datamodel.parent.get_unbinned()
            if self.datamodel.parent.db_only:
                in_db = unbinned in self.datamodel.parent.db_only
            else:
                in_db = unbinned in self.datamodel.parent.bintypes

            if self.mode == 'remote' and not in_db:
                raise marvin.core.exceptions.MarvinError(
                    "Bintype {0} for release {1} not available in remote database. Use "
                    "Marvin's local file access mode instead.".format(unbinned.name, self.release))

            return ModelCube(plateifu=self.plateifu, release=self.release,
                             bintype=self.datamodel.parent.get_unbinned(),
                             template=self.template,
                             mode=self.mode)



#!/usr/bin/env python3
# encoding: utf-8
#
# datamodel.py
#
# Created by José Sánchez-Gallego on 18 Sep 2016.
# Refactored by José Sánchez-Gallego on 6 Aug 2017.


from __future__ import absolute_import, division, print_function

import copy as copy_mod
import itertools
import os
import re

import astropy.table as table
import six
from astropy import units as u

from marvin.core.exceptions import MarvinError
from marvin.utils.datamodel import DataModelList
from marvin.utils.general.structs import FuzzyList, get_best_fuzzy


__ALL__ = ('DAPDataModelList', 'DAPDataModel', 'Bintype', 'Template', 'Property',
           'MultiChannelProperty', 'spaxel', 'datamodel', 'Channel')


spaxel = u.Unit('spaxel', represents=u.pixel, doc='A spectral pixel', parse_strict='silent')


class DAPDataModel(object):
    """A class representing a DAP datamodel.

    Parameters
    ----------
    release : str
        The DRP release this datamodel describes.
    bintypes : list
        A list of `.Bintype` instances that describe the bintypes available
        for this datamodel.
    templates : list
        A list of `.Template` instances that describe the stellar templates
        available for this datamodel.
    properties : list
        A list of `.Property` instances that describe the properties (maps)
        available in this datamodel.
    models : list
        A list of `.Model` instances that describe the model datacubes
        available for this datamodel.
    property_table : str
        The name of the DB table containing property data.
    default_bintype : str
        The name of the default bintype.
    default_template : str
        The name of the default template.
    default_binid : str
        The name of the default binid extension to use.
    db_only : list
        A list of bintypes that are accessible in the DB. If ``None``, assumes
        that all the bintypes are loaded in the DB.
    excluded_daptypes : list
        A list of DAPTYPE that do not exist and should be excluded from combinations
    default_mapset : list
        A list of maps to show in the web by default.
    default_mapmask : list
        A list of pixmask labels that define the bad mask.
    aliases : list
        A list of aliases for this datamodel.
    bitmask : dict
        A dictionary of `~marvin.utils.general.maskbit.Maskbit` objects.
    qual_flag : str
        The name of the quality bitmask flag. Must not include the ``MANGA_``
        prefix.

    """

    def __init__(self, release, bintypes=[], templates=[], properties=[], models=[],
                 default_template=None, default_bintype=None, property_table=None,
                 default_binid=None, aliases=[], bitmasks=None, db_only=[], default_mapset=None,
                 default_mapmask=None, qual_flag='MANGA_DAPQUAL', excluded_daptypes=[]):

        self.release = release
        self.bintypes = bintypes
        self.templates = templates
        self.db_only = db_only
        self.excluded_daptypes = excluded_daptypes

        self.aliases = aliases

        self.properties = PropertyList(properties, parent=self)
        self.models = ModelList(models, parent=self)

        self.bitmasks = bitmasks if bitmasks is not None else {}

        self.property_table = property_table

        # default plotting params
        self.default_mapmask = default_mapmask
        self.qual_flag = qual_flag
        self.default_mapset = self.get_default_mapset(default_mapset)
        self.default_plot_params = self.get_default_plot_params()

        # default bintypes/templates
        self._default_bintype = None
        self.default_bintype = default_bintype

        self._default_template = None
        self.default_template = default_template

        self.default_binid = default_binid
        if self.default_binid is not None:
            self.default_binid.parent = self

        assert len([bintype for bintype in self.bintypes if bintype.binned is False]) <= 1, \
            'a DAP datamodel can have only one unbinned bintype'

    def __repr__(self):

        return ('<DAPDataModel release={0!r}, n_bintypes={1}, n_templates={2}, n_properties={3}>'
                .format(self.release, len(self.bintypes), len(self.templates),
                        len(self.properties)))

    def copy(self):
        """Returns a copy of the datamodel."""

        return copy_mod.deepcopy(self)

    def __eq__(self, value):
        """Uses fuzzywuzzy to return the closest property match."""

        prop_names = [prop.name for prop in self.properties.extensions]
        model_names = [model.full() for model in self.models]

        # If the values matches exactly one of the property or model names,
        # we return that one.
        if value in prop_names:
            return self.properties.extensions[prop_names.index(value)]
        elif value in model_names:
            return self.models[model_names.index(value)]

        # Gets the best match for properties and models. Only returns a value
        # if only one of them gets a match.

        try:
            prop_best_match = self.properties[value]
        except ValueError:
            prop_best_match = None

        try:
            model_best_match = self.models[value]
        except ValueError:
            model_best_match = None

        if ((model_best_match is None and prop_best_match is None) or
                (model_best_match is not None and prop_best_match is not None)):
            raise ValueError('too ambiguous input {!r}'.format(value))
        elif model_best_match is not None:
            return model_best_match
        elif prop_best_match is not None:
            return prop_best_match

    def __contains__(self, value):

        try:
            match = self.__eq__(value)
            if match is None:
                return False
            else:
                return True
        except ValueError:
            return False

    def __getitem__(self, value):
        return self == value

    @property
    def default_bintype(self):
        """Returns the default bintype."""

        return self._default_bintype

    @default_bintype.setter
    def default_bintype(self, value):
        """Sets the default bintype."""

        for bintype in self.bintypes:
            if str(value) == bintype.name:
                value = bintype

        if value not in self.bintypes and value is not None:
            raise ValueError('{0!r} not found in list of bintypes.'.format(value))

        self._default_bintype = value

    @property
    def default_template(self):
        """Returns the default template."""

        return self._default_template

    @default_template.setter
    def default_template(self, value):
        """Sets the default template."""

        for template in self.templates:
            if str(value) == template.name:
                value = template

        if value not in self.templates and value is not None:
            raise ValueError('{0!r} not found in list of templates.'.format(value))

        self._default_template = value

    def get_unbinned(self):
        """Returns the unbinned bintype."""

        for bintype in self.bintypes:
            if bintype.binned is False:
                return bintype

        raise MarvinError('cannot find unbinned bintype for '
                          'DAP datamodel for release {0}'.format(self.release))

    def get_bintype(self, value=None):
        """Returns a bintype whose name matches ``value``.

        Returns the default one if ``value=None``.

        """

        if value is None:
            return self.default_bintype

        for bintype in self.bintypes:
            if bintype.name.upper() == str(value).upper():
                return bintype

        raise MarvinError('invalid bintype {0!r}'.format(value))

    def get_template(self, value=None):
        """Returns a template whose name matches ``value``.

        Returns the default one if ``value=None``.

        """

        if value is None:
            return self.default_template

        for template in self.templates:
            if template.name.upper() == str(value).upper():
                return template

        raise MarvinError('invalid template {0!r}'.format(value))

    def get_bintemps(self, default=False, db_only=False):
        """Returns a list of all combinations of bintype and template."""

        if default:
            return '{0}-{1}'.format(self.default_bintype.name, self.default_template.name)

        bins = [bintype.name for bintype in self.bintypes]
        temps = [template.name for template in self.templates]

        if db_only and self.db_only:
            bins = [b for b in bins if b in self.db_only]

        return ['-'.join(item) for item in list(itertools.product(bins, temps)) if '-'.join(item) not in self.excluded_daptypes]

    def get_properties(self, name):
        ''' Returns a list of properties of the given name '''
        assert isinstance(name, six.string_types), 'given name must be a string'
        prop_list = [prop for prop in self.properties if name.lower() in prop.name]
        return prop_list

    def get_channels(self, name, formatted=None):
        ''' Returns a list of channels of a given property name '''
        props = self.get_properties(name)
        channels = [prop.channel.formats['string'] if formatted else prop.channel for prop in props if prop.channel]
        return channels

    @staticmethod
    def get_default_mapset(default=None):
        ''' Return the default set of maps

        Syntax of a map name is parameter:channel

        Parameters:
            default (list):
                A list of default map names

        Returns:
            A list of map names

        '''
        if not default:
            default = ['stellar_vel', 'emline_gflux:ha_6564', 'specindex:d4000']

        assert len(default) <= 6, 'default maps must number less than six'
        return default

    def get_default_plot_params(self):
        ''' Return the default set of plotting params '''

        if not self.default_mapmask:
            self.default_mapmask = ['DONOTUSE']

        return self.set_base_params()

    def get_plot_params(self, prop=None):
        ''' Get one set of plotting params for a given property

        Parameters:
            prop (str):
                The name of the property
        Returns:
            The default set of plotting params for the given property
        '''

        defaults = self.get_default_plot_params()

        if 'vel' in prop:
            key = 'vel'
        elif 'sigma' in prop:
            key = 'sigma'
        else:
            key = 'default'
        return defaults[key]

    def set_base_params(self):
        ''' Set the base plotting param defaults '''

        return {'default': {'bitmasks': self.default_mapmask,
                            'cmap': 'linearlab',
                            'percentile_clip': [5, 95],
                            'symmetric': False,
                            'snr_min': 1},
                'vel': {'bitmasks': self.default_mapmask,
                        'cmap': 'RdBu_r',
                        'percentile_clip': [10, 90],
                        'symmetric': True,
                        'snr_min': None},
                'sigma': {'bitmasks': self.default_mapmask,
                          'cmap': 'inferno',
                          'percentile_clip': [10, 90],
                          'symmetric': False,
                          'snr_min': 1}}


class DAPDataModelList(DataModelList):
    """A dictionary of DAP datamodels."""

    base = {'DAPDataModel': DAPDataModel}


class PropertyList(FuzzyList):
    """Creates a list containing properties and their representation."""

    def __init__(self, the_list, parent=None):

        self.parent = parent
        self.extensions = []

        super(PropertyList, self).__init__([])

        for item in the_list:
            self.append(item, copy=True)

    def mapper(self, value):
        """A helper for the FuzzyList to determine the query value."""

        return value.full()

    def append(self, value, copy=True):
        """Appends with copy, and unpacking properties."""

        append_obj = value if copy is False else copy_mod.deepcopy(value)
        append_obj.parent = self.parent

        self.extensions.append(append_obj)

        if isinstance(append_obj, Property):
            super(PropertyList, self).append(append_obj)
        elif isinstance(append_obj, MultiChannelProperty):
            for prop in append_obj:
                super(PropertyList, self).append(prop)
        else:
            raise ValueError('invalid property of type {!r}'.format(type(append_obj)))

    def from_fits_extension(self, ext):
        """Returns the `.Property` whose FITS extension matches ``ext``."""

        matches = []

        for model in self:
            if model.fits_extension().lower() == ext.lower():
                matches.append(model)

        assert len(matches) == 1, 'more than one matches found. That should never happen.'

        return matches[0]

    @property
    def release(self):
        """The release of the parent `.DAPDataModel`."""

        return self.parent.release

    @property
    def qual_flag(self):
        """The qual_flag of the parent `.DAPDataModel`."""

        return self.parent.qual_flag

    def to_table(self, compact=True, pprint=False, description=False, max_width=1000):
        """Returns an astropy table with all the properties in this model.

        Parameters:
            compact (bool):
                If ``True``, groups extensions (multichannel properties) in one
                line. Otherwise, shows a row for each extension and channel.
            pprint (bool):
                Whether the table should be printed to screen using astropy's
                table pretty print.
            description (bool):
                If ``True``, an extra column with the description of the
                property will be added.
            max_width (int or None):
                A keyword to pass to ``astropy.table.Table.pprint()`` with the
                maximum width of the table, in characters.

        Returns:
            result (``astropy.table.Table``):
                If ``pprint=False``, returns an astropy table containing
                the name of the property, the channel (or channels, if
                ``compact=True``), whether the property has ``ivar`` or
                ``mask``, the units, and a description (if
                ``description=True``). Additonal information such as the
                bintypes, templates, release, etc. is included in
                the metadata of the table (use ``.meta`` to access them).

        """

        if compact:
            prop_table = table.Table(
                None, names=['name', 'channels', 'ivar', 'mask', 'unit', 'description',
                             'db_table', 'db_column', 'fits_extension'],
                dtype=['S20', 'S300', bool, bool, 'S20', 'S500', 'S20', 'S300', 'S20'])
        else:
            prop_table = table.Table(
                None, names=['name', 'channels', 'ivar', 'mask', 'unit', 'description',
                             'db_table', 'db_column', 'fits_extension'],
                dtype=['S20', 'S20', bool, bool, 'S20', 'S500', 'S20', 'S20', 'S20'])

        if self.parent:
            prop_table.meta['release'] = self.parent.release
            prop_table.meta['bintypes'] = self.parent.bintypes
            prop_table.meta['templates'] = self.parent.templates
            prop_table.meta['default_bintype'] = self.parent.default_bintype
            prop_table.meta['default_template'] = self.parent.default_template

        if compact:
            iterable = self.extensions
        else:
            iterable = self

        for prop in iterable:
            if isinstance(prop, MultiChannelProperty):
                channel = ', '.join([str(channel) for channel in prop.channels])
                units = [pp.unit.to_string() for pp in prop]
                unit = units[0] if len(set(units)) == 1 else 'multiple'
                dbcolumn = ', '.join(prop.db_columns())
            else:
                channel = '' if not prop.channel else prop.channel
                unit = prop.unit.to_string()
                dbcolumn = prop.db_column()

            prop_table.add_row((prop.name, channel, prop.ivar, prop.mask, unit, prop.description,
                                prop.db_table, dbcolumn, prop.fits_extension()))

        if not description:
            prop_table.remove_column('description')

        if pprint:
            prop_table.pprint(max_width=max_width, max_lines=1e6)
            return

        return prop_table

    def write_csv(self, filename=None, path=None, overwrite=None, **kwargs):
        ''' Write the property datamodel to a CSV '''

        release = self.parent.aliases[0].lower().replace('-', '')

        if not filename:
            filename = 'dapprops_dm_{0}.csv'.format(release)

        if not path:
            path = os.path.join(os.getenv("MARVIN_DIR"), 'docs', 'sphinx', '_static')

        fullpath = os.path.join(path, filename)
        table = self.to_table(**kwargs)
        table.write(fullpath, format='csv', overwrite=overwrite)


class ModelList(FuzzyList):
    """Creates a list containing models and their representation."""

    def __init__(self, the_list, parent=None):

        self.parent = parent

        super(ModelList, self).__init__([])

        for item in the_list:
            self.append(item, copy=True)

    def mapper(self, value):
        """A helper for the FuzzyList to determine the query value."""

        return value.full()

    def append(self, value, copy=True):
        """Appends with copy."""

        append_obj = value if copy is False else copy_mod.deepcopy(value)
        append_obj.parent = self.parent

        if isinstance(append_obj, Model):
            super(ModelList, self).append(append_obj)
        else:
            raise ValueError('invalid model of type {!r}'.format(type(append_obj)))

    @property
    def release(self):
        """The release of the parent `.DAPDataModel`."""

        return self.parent.release

    @property
    def qual_flag(self):
        """The qual_flag of the parent `.DAPDataModel`."""

        return self.parent.qual_flag

    def from_fits_extension(self, extension):
        """Returns the `.Model` whose FITS extension matches ``extension``."""

        matches = []

        for model in self:
            if model.fits_extension().lower() == extension.lower():
                matches.append(model)

        assert len(matches) == 1, 'more than one matches found. That should never happen.'

        return matches[0]

    def to_table(self, pprint=False, description=False, max_width=1000):
        """Returns an astropy table with all the models in this datamodel.

        Parameters:
            pprint (bool):
                Whether the table should be printed to screen using astropy's
                table pretty print.
            description (bool):
                If ``True``, an extra column with the description of the
                model will be added.
            max_width (int or None):
                A keyword to pass to ``astropy.table.Table.pprint()`` with the
                maximum width of the table, in characters.

        Returns:
            result (``astropy.table.Table``):
                If ``pprint=False``, returns an astropy table containing
                the name of the model, whether the property has ``ivar`` or
                ``mask``, the units, and a description (if
                ``description=True``). Additonal information such as the
                bintypes, templates, release, etc. is included in
                the metadata of the table (use ``.meta`` to access them).

        """

        model_table = table.Table(
            None, names=['name', 'ivar', 'mask', 'unit', 'description',
                         'db_table', 'db_column', 'fits_extension'],
            dtype=['S20', bool, bool, 'S20', 'S500', 'S20', 'S20', 'S20'])

        if self.parent:
            model_table.meta['release'] = self.parent.release
            model_table.meta['bintypes'] = self.parent.bintypes
            model_table.meta['templates'] = self.parent.templates
            model_table.meta['default_bintype'] = self.parent.default_bintype
            model_table.meta['default_template'] = self.parent.default_template

        for model in self:
            unit = model.unit.to_string()

            model_table.add_row((model.name,
                                 model._extension_ivar is not None,
                                 model._extension_mask is not None,
                                 unit,
                                 model.description,
                                 model.db_table,
                                 model.db_column(),
                                 model.fits_extension()))

        if not description:
            model_table.remove_column('description')

        if pprint:
            model_table.pprint(max_width=max_width, max_lines=1e6)
            return

        return model_table

    def write_csv(self, filename=None, path=None, overwrite=None, **kwargs):
        ''' Write the datamodel to a CSV '''

        release = self.parent.aliases[0].lower().replace('-', '')

        if not filename:
            filename = 'dapmodels_dm_{0}.csv'.format(release)

        if not path:
            path = os.path.join(os.getenv("MARVIN_DIR"), 'docs', 'sphinx', '_static')

        fullpath = os.path.join(path, filename)
        table = self.to_table(**kwargs)
        table.write(fullpath, format='csv', overwrite=overwrite)


class Bintype(object):
    """A class representing a bintype."""

    def __init__(self, bintype, binned=True, n=None, description=''):

        self.name = bintype
        self.binned = binned
        self.n = n
        self.description = description

    def __repr__(self):

        return '<Bintype {0!r}, binned={1!r}>'.format(self.name, self.binned)

    def __str__(self):

        return self.name

    def __eq__(self, other):

        return self.name == other or super(Bintype, self) == other


class Template(object):
    """A class representing a template."""

    def __init__(self, template, n=None, description=''):

        self.name = template
        self.n = n
        self.description = description

    def __repr__(self):

        return '<Template {0!r}>'.format(self.name)

    def __str__(self):

        return self.name

    def __eq__(self, other):

        return self.name == other or super(Template, self) == other


class Property(object):
    """A class representing a DAP property.

    Parameters:
        name (str):
            The name of the property.
        channel (:class:`Channel` object or None):
            The channel associated to the property, if any.
        ivar (bool):
            Whether the property has an inverse variance measurement.
        mask (bool):
            Whether the property has an associated mask.
        unit (astropy unit or None):
            The unit for this channel. If not defined, the unit from the
            ``channel`` will be used.
        scale (float or None):
            The scaling factor for the property. If not defined, the scaling
            factor from the ``channel`` will be used.
        formats (dict):
            A dictionary with formats that can be used to represent the
            property. Default ones are ``latex`` and ``string``.
        parent (:class:`DAPDataModel` object or None):
            The associated :class:`DAPDataModel` object. Usually it is set to
            ``None`` and populated when the property is added to the
            ``DAPDataModel`` object.
        binid (:class:`Property` object or None):
            The ``binid`` :class:`Property` object associated with this
            propety. If not set, assumes the `.DAPDataModel` ``default_binid``.
        pixmask_flag : str
            The name of the pixmask flag. Should be the full name, including
            the ``MANGA_`` prefix.
        description (str):
            A description of the property.

    """

    def __init__(self, name, channel=None, ivar=False, mask=False, unit=None,
                 scale=1, formats={}, parent=None, binid=None,
                 pixmask_flag='MANGA_DAPPIXMASK', description=''):

        self.name = name
        self.channel = copy_mod.deepcopy(channel)

        self.ivar = ivar
        self.mask = mask

        self.pixmask_flag = pixmask_flag

        self.formats = formats

        if unit is not None:
            self.unit = u.CompositeUnit(scale, unit.bases, unit.powers)
        elif unit is None and self.channel is None:
            self.unit = u.dimensionless_unscaled
        else:
            self.unit = self.channel.unit

        self._binid = binid

        # Makes sure the channel shares the units and scale
        if self.channel:
            self.channel.unit = self.unit

        self.description = description

        self._parent = None
        self.parent = parent

        self._binid = copy_mod.deepcopy(binid)
        if self._binid is not None:
            self._binid.parent = self.parent

    @property
    def parent(self):
        """Returns the parent for this property."""

        return self._parent

    @parent.setter
    def parent(self, value):
        """Sets the parent."""

        assert value is None or isinstance(value, DAPDataModel), 'value must be a DAPDataModel'

        self._parent = value

        if hasattr(self, '_binid') and self._binid is not None:
            self._binid.parent = value

    def full(self, web=None, db=None):
        """Returns the name + channel string."""

        if self.channel:
            channel_name = self.channel.db_name if db else self.channel.name
            if web:
                return self.name + ':' + channel_name
            else:
                return self.name + '_' + channel_name

        return self.name

    @property
    def binid(self):
        """Returns the binid property associated with this property."""

        if self.name == 'binid':
            raise MarvinError('binid has not associated binid (?!)')

        assert self.parent is not None, 'a parent needs to be defined to get an associated binid.'

        if self._binid is None:
            return self.parent.default_binid

        return self._binid

    def has_ivar(self):
        """Returns True if the property has an ivar extension."""

        return self.ivar is not False

    def has_mask(self):
        """Returns True if the property has an mask extension."""

        return self.mask is not False

    def db_column(self, ext=None):
        """Returns the name of the DB column containing this property."""

        assert ext is None or ext in ['ivar', 'mask'], 'invalid extension'

        if ext is None:
            return self.full(db=True)

        if ext == 'ivar':
            assert self.ivar is True, 'no ivar for property {0!r}'.format(self.full(db=True))
            return self.name + '_ivar' + \
                ('_{0}'.format(self.channel.db_name) if self.channel else '')

        if ext == 'mask':
            assert self.mask is True, 'no mask for property {0!r}'.format(self.full(db=True))
            return self.name + '_mask' + \
                ('_{0}'.format(self.channel.db_name) if self.channel else '')

    def __repr__(self):

        return '<Property {0!r}, channel={2!r}, release={1!r}, unit={3!r}>'.format(
            self.name, self.parent.release if self.parent else None,
            self.channel.name if self.channel else 'None', self.unit.to_string())

    def __str__(self):

        return self.full()

    @property
    def model(self):
        ''' The ModelClass the property belongs to '''

        return self.parent.property_table

    @property
    def db_table(self):
        """The DB table to use to retrieve this property."""

        assert self.parent is not None, 'parent DAPDataModel is not set for this property.'

        return self.parent.property_table.lower()

    def fits_extension(self):
        ''' The FITS extension this property belongs to '''

        ext = self.name.upper()
        if self.channel:
            channel_num = self.channel.idx
            ext = '{0}_{1}'.format(ext, channel_num)
        return ext

    def to_string(self, mode='string', include_channel=True):
        """Return a string representation of the channel."""

        if mode == 'latex':

            if mode in self.formats:
                latex = self.formats[mode]
            else:
                latex = self.to_string(include_channel=False)

            if self.channel and include_channel:
                latex = latex + ' ' + self.channel.to_string('latex')

            return latex

        else:

            if mode in self.formats:
                string = self.formats[mode]
            else:
                string = self.name

            if self.channel is None or include_channel is False:
                return string
            else:
                return string + ': ' + self.channel.to_string(mode=mode)


class MultiChannelProperty(list):
    """A class representing a list of channels for the same property.

    Parameters:
        name (str):
            The name of the property.
        channels (list of :class:`Channel` objects):
            The channels associated to the property.
        ivar (bool):
            Whether the properties have an inverse variance measurement.
        mask (bool):
            Whether the properties have an associated mask.
        unit (astropy unit or None):
            The unit for these channels. If set, it will override any unit
            defined in the individual channels.
        scale (float):
            The scaling factor for these channels. If set, it will override
            any unit defined in the individual channels.
        formats (dict):
            A dictionary with formats that can be used to represent the
            property. Default ones are ``latex`` and ``string``.
        parent (:class:`DAPDataModel` object or None):
            The associated :class:`DAPDataModel` object. Usually it is set to
            ``None`` and populated when the property is added to the
            ``DAPDataModel`` object.
        binid (:class:`Property` object or None):
            The ``binid`` `.Property` object to be associated to all the
            propeties in this `.MultiChannelProperty`.
        description (str):
            A description of the property.
        kwargs (dict):
            Arguments to be passed to each ``Property`` on initialisation.

    """

    def __init__(self, name, channels=[], unit=None, scale=1, binid=None, **kwargs):

        self.name = name

        self.ivar = kwargs.get('ivar', False)
        self.mask = kwargs.get('mask', False)
        self.description = kwargs.pop('description', '')

        self._parent = None
        self.parent = kwargs.get('parent', None)

        self._binid = binid

        self_list = self._generate_prop_list(channels, unit=unit, scale=scale, **kwargs)
        # self_list = []
        # for ii, channel in enumerate(channels):
        #     this_unit = unit if not isinstance(unit, (list, tuple)) else unit[ii]
        #     this_scale = scale if not isinstance(scale, (list, tuple)) else scale[ii]
        #     self_list.append(self._generate_property(channel, unit=this_unit, scale=this_scale, **kwargs))

        list.__init__(self, self_list)

    @property
    def parent(self):
        """Returns the parent for this MultiChannelProperty."""

        return self._parent

    @parent.setter
    def parent(self, value):
        """Sets parent for the instance and all listed Property objects."""

        assert value is None or isinstance(value, DAPDataModel), 'value must be a DAPDataModel'

        self._parent = value

        for prop in self:
            prop.parent = value

    @property
    def channels(self):
        """Returns a list of channels."""

        return [item.channel for item in self]

    def __getitem__(self, value):
        """Uses fuzzywuzzy to get a channel."""

        if not isinstance(value, six.string_types):
            return super(MultiChannelProperty, self).__getitem__(value)

        best_match = get_best_fuzzy(value, self.channels)

        return super(MultiChannelProperty, self).__getitem__(self.channels.index(best_match))

    def __repr__(self):

        return '<MultiChannelProperty {0!r}, release={1!r}, channels={2!r}>'.format(
            self.name, self.parent.release if self.parent else None,
            [channel.name for channel in self.channels])

    @property
    def db_table(self):
        ''' Returns the db table this belongs to '''

        return self.parent.property_table.lower()

    def db_columns(self):
        ''' Returns a list of db columns for this MultiChannelProperty '''

        return [item.db_column() for item in self]

    def fits_extension(self):
        ''' Returns the FITS extension this belongs to '''

        return self.name.upper()

    def _generate_prop_list(self, channels, unit=None, scale=1, **kwargs):
        ''' Generate a new Property list from a list of Channels '''
        self_list = []
        for ii, channel in enumerate(channels):
            this_unit = unit if not isinstance(unit, (list, tuple)) else unit[ii]
            this_scale = scale if not isinstance(scale, (list, tuple)) else scale[ii]
            self_list.append(self._generate_property(
                channel, unit=this_unit, scale=this_scale, **kwargs))
        return self_list

    def _generate_property(self, channel, unit=None, scale=1, **kwargs):
        ''' Generate a new Property '''

        channel_description = self.description + ' Channel = {0}.'.format(channel.to_string())
        new_prop = Property(self.name, channel=channel,
                            unit=unit, scale=scale,
                            binid=self._binid, description=channel_description, **kwargs)
        return new_prop

    def append_channel(self, channel, unit=None, scale=1, at_index=None, **kwargs):
        ''' Append a Channel to the MultiChannelProperty list
        
        Appends a channel to the existing list of channel.  Can either insert the
        channel at a specified index or at the end of the list (default).  If an index
        is specified, will automatically reindex the channels to have the proper idx.    

        Parameters:
            channel (Channel):
                The channel to append
            at_index (int):
                The list index at which to insert the channel.  Default is None.
            unit (astropy unit or None):
                The unit for these channels. If set, it will override any unit
                defined in the individual channels.
            scale (float):
                The scaling factor for these channels. If set, it will override
                any unit defined in the individual channels.
            kwargs (dict):
                Arguments to be passed to each ``Property`` on initialisation.
        '''

        new_prop = self._generate_property(channel, unit=unit, scale=scale, **kwargs)
        n_item = len(self)
        idx = at_index if at_index else n_item
        self.insert(idx, new_prop)
        # if an insert index is given, then automatically reindex
        if at_index is not None:
            self.reindex_channels()

    def reindex_channels(self):
        ''' Reindex the Channels in the Property list '''

        for idx, channel in enumerate(self.channels):
            channel.idx = idx

    def update_channels(self, channels, unit=None, scale=1, **kwargs):
        ''' Update the Channels in the MultiChannelProperty

        Parameters:
            channels (list):
                new list of Channels to replace existing ones
            unit (astropy unit or None):
                The unit for these channels. If set, it will override any unit
                defined in the individual channels.
            scale (float):
                The scaling factor for these channels. If set, it will override
                any unit defined in the individual channels.
            kwargs (dict):
                Arguments to be passed to each ``Property`` on initialisation.
        '''
        self_list = self._generate_prop_list(channels, unit=unit, scale=scale, **kwargs)
        list.__init__(self, self_list)


class Channel(object):
    """A class representing a channel in a property.

    Parameters:
        name (str):
            The channel name.
        unit (astropy unit or None):
            The unit for this channel.
        scale (float):
            The scaling factor for the channel.
        formats (dict):
            A dictionary with formats that can be used to represent the
            channel. Default ones are ``latex`` and ``string``.
        idx (int):
            The index of the channel in the MAPS file extension.
        db_name (str or None):
            The name of this channel in the database. If None, ``name`` will
            be used.
        description (str):
            A description for the channel.

    """

    def __init__(self, name, unit=u.dimensionless_unscaled, scale=1, formats={},
                 idx=None, db_name=None, description=''):

        self.name = name
        self.unit = u.CompositeUnit(scale, unit.bases, unit.powers)
        self.formats = formats
        self.idx = idx
        self.db_name = db_name or self.name
        self.description = description

    def to_string(self, mode='string'):
        """Return a string representation of the channel."""

        if mode == 'latex':
            if 'latex' in self.formats:
                latex = self.formats['latex']
                latex = re.sub(r'forb{(.+)}', r'lbrack\\textrm{\1}\\rbrack', latex)
            else:
                latex = self.to_string().replace(' ', '\\ ')
            return latex
        elif mode is not None and mode in self.formats:
            return self.formats[mode]
        else:
            return self.name

    def __repr__(self):

        return '<Channel {0!r} unit={1!r}>'.format(self.name, self.unit.to_string())

    def __str__(self):

        return self.name


class Model(object):
    """Represents a extension in the DAP logcube file.

    Parameters:
        name (str):
            The model name. This is the internal name that Marvin will use for
            this model. It is different from the ``extension_name`` parameter,
            which must be identical to the extension name of the model.
        extension_name (str):
            The FITS extension containing this model.
        extension_wave (str):
            The FITS extension containing the wavelength for this model.
        extension_ivar (str or None):
            The extension that contains the inverse variance associated with
            this model, if any.
        extension_mask (str or None):
            The extension that contains the mask associated with this model,
            if any.
        channels (list):
            The channels associated with this model (probably only used
            for binid).
        unit (astropy unit or None):
            The unit for this model.
        scale (float):
            The scaling factor for the values of the model.
        formats (dict):
            A dictionary with formats that can be used to represent the
            model. Default ones are ``latex`` and ``string``.
        parent (:class:`DAPDataModel` object or None):
            The associated :class:`DAPDataModel` object. Usually it is set to
            ``None`` and populated when the model is added to the
            ``DAPDataModel`` object.
        binid (:class:`Property` object or None):
            The ``binid`` :class:`Property` object associated with this
            model. If not set, assumes the `.DAPDataModel` ``default_binid``.
        pixmask_flag : str
            The name of the pixmask flag. Should be the full name, including
            the ``MANGA_`` prefix.
        description (str):
            A description for the model.
        db_table (str):
            The database table the model belongs to.

    """

    def __init__(self, name, extension_name, extension_wave=None,
                 extension_ivar=None, extension_mask=None, channels=[],
                 unit=u.dimensionless_unscaled, scale=1, formats={},
                 parent=None, binid=None, description='',
                 db_table='modelspaxel', pixmask_flag='MANGA_DAPSPECMASK'):

        self.name = name

        self._extension_name = extension_name
        self._extension_wave = extension_wave
        self._extension_ivar = extension_ivar
        self._extension_mask = extension_mask

        self.channels = channels

        self.unit = u.CompositeUnit(scale, unit.bases, unit.powers)

        self.formats = formats
        self.description = description
        self.db_table = db_table

        self._binid = binid

        self._parent = None
        self.parent = parent

        self.pixmask_flag = pixmask_flag

        self._binid = copy_mod.deepcopy(binid)
        if self._binid is not None:
            self._binid.parent = self.parent

    @property
    def parent(self):
        """Returns the parent for this model."""

        return self._parent

    @parent.setter
    def parent(self, value):
        """Sets parent."""

        assert value is None or isinstance(value, DAPDataModel), 'value must be a DAPDataModel'

        self._parent = value

        if self._binid is not None:
            self._binid.parent = value

    def full(self):
        """Returns the name + channel string."""

        return self.name

    def has_ivar(self):
        """Returns True if the datacube has an ivar extension."""

        return self._extension_ivar is not None

    def has_mask(self):
        """Returns True if the datacube has an mask extension."""

        return self._extension_mask is not None

    @property
    def binid(self):
        """Returns the binid property associated with this property."""

        if self.name == 'binid':
            raise MarvinError('binid has not associated binid (?!)')

        assert self.parent is not None, 'a parent needs to be defined to get an associated binid.'

        if self._binid is None:
            return self.parent.default_binid

        return self._binid

    def fits_extension(self, ext=None):
        """Returns the FITS extension name."""

        assert ext is None or ext in ['ivar', 'mask'], 'invalid extension'

        if ext is None:
            return self._extension_name.upper()

        elif ext == 'ivar':
            if not self.has_ivar():
                raise MarvinError('no ivar extension for datacube {0!r}'.format(self.full()))
            return self._extension_ivar.upper()

        elif ext == 'mask':
            if not self.has_mask():
                raise MarvinError('no mask extension for datacube {0!r}'.format(self.full()))
            return self._extension_mask

    def db_column(self, ext=None):
        """Returns the name of the DB column containing this datacube."""

        return self.fits_extension(ext=ext).lower()

    def __repr__(self):

        return '<Model {!r}, release={!r}, unit={!r}>'.format(
            self.name, self.parent.release if self.parent else None, self.unit.to_string())

    def __str__(self):

        return self.full()

    def to_string(self, mode='string', include_channel=True):
        """Return a string representation of the channel."""

        if mode == 'latex':

            if mode in self.formats:
                latex = self.formats[mode]
            else:
                latex = self.to_string(include_channel=False)

            if self.channel and include_channel:
                latex = latex + ' ' + self.channel.to_string('latex')

            return latex

        else:

            if mode in self.formats:
                string = self.formats[mode]
            else:
                string = self.name

            if self.channel is None or include_channel is False:
                return string
            else:
                return string + ': ' + self.channel.to_string(mode=mode)


# functions to aid in updating datamodel
def reindex_channels(channels, names=None, indices=None, starting_idx=None):
    ''' Reindex a list of channels
    
    Parameters:
        channels (list):
            List of Channel objects to reindex
        names (list):
            A list of channel names to reindex by
        indices (list):
            A list of integers to reindex by
        starting_idx (int):
            An starting index to offset by when reindexing
    
    Returns:
        A new sorted and reindexed list of Channels

    '''

    # sort the channels by new names or indices
    if names:
        channels = sorted(channels, key=lambda t: names.index(t.name))

    if indices:
        channels = [i for i, __ in sorted((zip(channels, indices)),key=lambda t:t[1])]

    # update the channel indexes
    for i, channel in enumerate(channels):
        idx = i + starting_idx if starting_idx else i
        channel.idx = idx

    return channels
    



#!/usr/bin/env python
# encoding: utf-8

'''
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-02-17 17:46:57
    Last Modified On: 2016-02-17 17:46:57 by Brian

'''
from __future__ import print_function
from __future__ import division
import os
from brain.api.base import BrainBaseView
from brain.utils.general import build_routemap
from marvin import config
from marvin.api import ArgValidator, set_api_decorators
from flask import current_app, abort


arg_validate = ArgValidator(urlmap=None)


class BaseView(BrainBaseView):
    '''Super Class for all API Views to handle all global API items of interest'''
    decorators = set_api_decorators()

    def add_config(self):
        utahconfig = {'utahconfig': {'mode': config.mode, 'release': config.release}}
        self.update_results(utahconfig)

    def _pop_args(self, kwargs, arglist=None):
        if arglist:
            arglist = [arglist] if not isinstance(arglist, (list, tuple)) else arglist

            for item in arglist:
                tmp = kwargs.pop(item, None)

        return kwargs

    def before_request(self, *args, **kwargs):

        # try to get a local version of the urlmap for the arg_validator
        if not arg_validate.urlmap:
            urlmap = build_routemap(current_app)
            config.urlmap = urlmap
            arg_validate.urlmap = urlmap

        super(BaseView, self).before_request(*args, **kwargs)

        self.check_release(self._release)

    def set_allowed_releases(self):
        ''' set the supported API / web MaNGA releases '''
        web_releases = current_app.config.get('WEB_RELEASES', None)
        public = config.access == 'public' or os.environ.get('PUBLIC_SERVER', None) is True
        self.allowed_releases = config.get_allowed_releases(public=public, web_releases=web_releases)

    def check_release(self, release):
        ''' check a release against the list of supported API releases '''
        self.set_allowed_releases()

        if release not in self.allowed_releases.keys():
            msg = "Error: data release {0} no longer supported by the Marvin API. " \
                  "Update to a later MPL or use Marvin's local file access mode " \
                  "instead.".format(release)
            abort(409, description=msg)




#!/usr/bin/env python
# encoding: utf-8

# Created by Brian Cherinka on 2016-02-18 18:04:14
# Licensed under a 3-clause BSD license.

# Revision History:
#     Initial Version: 2016-02-18 18:04:14 by Brian Cherinka
#     Last Modified On: 2016-02-18 18:04:14 by Brian

from __future__ import print_function
from __future__ import division
import numpy as np
import flask
import jinja2


# If the filter is to return HTML code and you don't want it autmatically
# escaped, return the value as "return Markup(value)".

jinjablue = flask.Blueprint('jinja_filters', __name__)

# Ref: http://stackoverflow.com/questions/12288454/how-to-import-custom-jinja2-filters-from-another-file-and-using-flask


@jinja2.contextfilter
@jinjablue.app_template_filter()
def make_token(context, value, group):
    ''' Make a keyword string for query parameter dropdown live search '''
    tokstring = ', '.join([group, value._joinedname])
    return tokstring


@jinja2.contextfilter
@jinjablue.app_template_filter()
def filtergaltype(context, value):
    ''' Parse plateifu or mangaid into better form '''
    if value == 'plateifu':
        return 'Plate-IFU'
    elif value == 'mangaid':
        return 'MaNGA-ID'


@jinja2.contextfilter
@jinjablue.app_template_filter()
def filternsa(context, value):
    ''' Parse plateifu or mangaid into better form '''

    newvalue = value.replace('elpetro_absmag_g_r', 'Abs. g-r').\
        replace('elpetro_absmag_u_r', 'Abs. u-r').\
        replace('elpetro_absmag_i_z', 'Abs. i-z')
    return newvalue


@jinja2.contextfilter
@jinjablue.app_template_filter()
def filternsaval(context, value, key):
    ''' Parse plateifu or mangaid into better form '''

    if type(value) == list:
        newvalue = ', '.join([str(np.round(v, 4)) for v in value])
    elif isinstance(value, (float, np.floating)):
        newvalue = np.round(value, 4)
    else:
        newvalue = value

    return newvalue


@jinja2.contextfilter
@jinjablue.app_template_filter()
def allclose(context, value, newvalue):
    ''' Do a numpy allclose comparison between the two values '''
    try:
        return np.allclose(float(value), float(newvalue), 1e-7)
    except Exception as e:
        return False


@jinja2.contextfilter
@jinjablue.app_template_filter()
def prettyFlag(context, value):
    ''' Pretty print bit mask and flags '''
    name, bit, flags = value
    return '{0}: {1} - {2}'.format(name, bit, ', '.join(flags))


@jinja2.contextfilter
@jinjablue.app_template_filter()
def qaclass(context, value):
    ''' Return an alert indicator based on quality flags '''
    name, bit, flags = value
    isgood = ['VALIDFILE'] == flags or [] == flags
    iscrit = 'CRITICAL' in flags
    out = 'success' if isgood else 'danger' if iscrit else 'warning'
    text = 'Good' if isgood else 'DO NOT USE' if iscrit else 'Warning'
    return out, text


@jinja2.contextfilter
@jinjablue.app_template_filter()
def targtype(context, value):
    ''' Return the MaNGA target type based on what bit is set '''
    # names = value.get('names', None)
    # namelabel = ', '.join(names)
    # out = namelabel.replace('MNGTRG1', 'Galaxy').replace('MNGTRG2', 'Stellar').replace('MNGTRG3', 'Ancillary')
    out = 'Galaxy' if '1' in value else 'Ancillary' if '3' in value else 'Stellar'
    return out


@jinja2.contextfilter
@jinjablue.app_template_filter()
def split(context, value, delim=None):
    '''Split a string based on a delimiter'''
    if not delim:
        delim = ' '
    return value.split(delim) if value else None


@jinja2.contextfilter
@jinjablue.app_template_filter()
def striprelease(context, value):
    '''Strip and trim and lowercase a release string'''
    value = value.strip().replace('-', '').lower()
    return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-10-30
# @Filename: map.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by:   andrews
# @Last modified time: 2019-11-22 11:11:71


from __future__ import absolute_import, division, print_function

import operator
import os
import warnings
from copy import deepcopy

import astropy.units as units
import numpy as np

import marvin
import marvin.api.api
import marvin.core.exceptions
import marvin.core.marvin_pickle
import marvin.utils.general
import marvin.utils.plot.map
from marvin.utils.datamodel.dap.base import Property
from marvin.utils.general.general import add_doc
from marvin.tools.mixins.caching import CacheMixIn

from .base_quantity import QuantityMixIn


try:
    import sqlalchemy
except ImportError:
    sqlalchemy = None


class Map(units.Quantity, QuantityMixIn):
    """Describes 2D array object with addtional features.

    A general-use `~astropy.units.Quantity`-powered 2D array with features
    such as units, inverse variance, mask, quick acces to statistics
    methods, and plotting. While `.Map` can be used for any 2D array, it is
    mostly intended to represent an extension in a DAP MAPS file (normally
    known in MaNGA as a map). Unlike a `~marvin.tools.maps.Maps` object,
    which contains all the information from a DAP MAPS file, this class
    represents only one of the multiple 2D maps contained
    within. For instance, `~marvin.tools.maps.Maps` may contain emission line
    maps for multiple channels. A `.Map` would be, for example, the map for
    ``emline_gflux`` and channel ``ha_6564``.

    A `.Map` is normally initialised from a `~marvin.tools.maps.Maps` by
    calling the `~marvin.tools.maps.Maps.getMap` method. It can be
    initialialised directly using the `.Map.from_maps` classmethod.

    Parameters:
        array (array-like):
            The 2-D array contianing the spectrum.
        unit (astropy.unit.Unit, optional):
            The unit of the spectrum.
        scale (float, optional):
            The scale factor of the spectrum value.
        ivar (array-like, optional):
            The inverse variance array for ``value``.
        mask (array-like, optional):
            The mask array for ``value``.
        binid (array-like, optional):
            The associated binid map.
        pixmask_flag (str):
            The maskbit flag to be used to convert from mask bits to labels
            (e.g., MANGA_DAPPIXMASK).

    """

    def __new__(cls, array, unit=None, scale=1, ivar=None, mask=None,
                binid=None, pixmask_flag=None, dtype=None, copy=True):

        if scale is not None:
            unit = units.CompositeUnit(unit.scale * scale, unit.bases, unit.powers)

        obj = units.Quantity(np.array(array), unit=unit, dtype=dtype, copy=copy)
        obj = obj.view(cls)
        obj._set_unit(unit)

        obj._maps = None
        obj._datamodel = None

        obj.ivar = np.array(ivar) if ivar is not None else None
        obj.mask = np.array(mask) if mask is not None else None
        obj.binid = np.array(binid) if binid is not None else None

        obj.pixmask_flag = pixmask_flag

        return obj

    def __repr__(self):

        if self.datamodel is not None:
            full = self.datamodel.full()
        else:
            full = None

        return ('<Marvin Map (property={0!r})>\n{1} {2}'.format(full, self.value,
                                                                self.unit.to_string()))

    def __getitem__(self, sl):

        new_obj = super(Map, self).__getitem__(sl)

        if type(new_obj) is not type(self):
            new_obj = self._new_view(new_obj)

        new_obj._set_unit(self.unit)

        new_obj.ivar = self.ivar.__getitem__(sl) if self.ivar is not None else self.ivar
        new_obj.mask = self.mask.__getitem__(sl) if self.mask is not None else self.mask
        new_obj.binid = self.binid.__getitem__(sl) if self.binid is not None else self.binid

        return new_obj

    def __deepcopy__(self, memo):

        new_map = Map(array=deepcopy(self.value, memo),
                      unit=deepcopy(self.unit, memo),
                      ivar=deepcopy(self.ivar, memo),
                      mask=deepcopy(self.mask, memo),
                      binid=deepcopy(self.binid, memo))

        new_map._maps = deepcopy(self._maps, memo)

        # TODO: temporary fix because the datamodel does deepcopy. Fix.
        new_map._datamodel = self._datamodel

        new_map.manga_target1 = deepcopy(self.manga_target1)
        new_map.manga_target2 = deepcopy(self.manga_target2)
        new_map.manga_target3 = deepcopy(self.manga_target3)
        new_map.target_flags = deepcopy(self.target_flags)

        new_map.pixmask_flag = deepcopy(self.pixmask_flag)

        return new_map

    def __array_finalize__(self, obj):

        if obj is None:
            return

        self._datamodel = getattr(obj, '_datamodel', None)
        self._maps = getattr(obj, '_maps', None)

        self.ivar = getattr(obj, 'ivar', None)
        self.mask = getattr(obj, 'mask', None)
        self.binid = getattr(obj, 'binid', None)
        self.pixmask_flag = getattr(obj, 'pixmask_flag', None)

        self._set_unit(getattr(obj, 'unit', None))

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):

        ivar = self._ivar_ufunc(ufunc, *inputs, **kwargs)

        return EnhancedMap(
            value=ufunc(self.value),
            ivar=ivar,
            mask=self.mask,
            unit=self.unit,  # unit
            datamodel=self._datamodel,
            pixmask_flag=self.pixmask_flag,
            copy=True,
        )

    def _ivar_ufunc(self, ufunc, *inputs, **kwargs):

        if ufunc in [np.log10]:
            compute_ivar = getattr(self, '_{}_ivar'.format(ufunc.__name__))
            ivar = compute_ivar(ivar=inputs[0].ivar, value=inputs[0].value, **kwargs)

        else:
            raise NotImplementedError('np.{0} is not implemented for {1}.'
                                      .format(ufunc.__name__, self.__class__.__name__))

        return ivar

    def getMaps(self):
        """Returns the associated `~marvin.tools.maps.Maps`."""

        if not self._maps:
            raise ValueError('this Map is not linked to any MAPS.')

        return self._maps

    @property
    def datamodel(self):
        """Returns the associated `~marvin.utils.datamodel.dap.Property`."""

        if not self._datamodel:
            raise ValueError('this Map does not have an associated Property.')

        return self._datamodel

    @classmethod
    def from_maps(cls, maps, prop, dtype=None, copy=True):
        """Initialise a `.Map` from a `.~marvin.tools.maps.Maps`."""

        import marvin.tools.maps

        assert isinstance(maps, marvin.tools.maps.Maps)
        assert isinstance(prop, Property)

        maps = maps
        datamodel = maps.datamodel

        assert prop.full() in datamodel, 'failed sanity check. Property does not match.'

        if maps.data_origin == 'file':
            value, ivar, mask = cls._get_map_from_file(maps, prop)
        elif maps.data_origin == 'db':
            value, ivar, mask = cls._get_map_from_db(maps, prop)
        elif maps.data_origin == 'api':
            value, ivar, mask = cls._get_map_from_api(maps, prop)

        # Gets the binid array for this property.
        if prop.name != 'binid':
            binid = maps.get_binid(prop)
        else:
            binid = None

        unit = prop.unit

        obj = cls(value, unit=unit, ivar=ivar, mask=mask, binid=binid,
                  dtype=dtype, copy=copy)

        obj._datamodel = prop
        obj._maps = maps

        obj.manga_target1 = maps.manga_target1
        obj.manga_target2 = maps.manga_target2
        obj.manga_target3 = maps.manga_target3
        obj.target_flags = maps.target_flags

        obj.pixmask_flag = prop.pixmask_flag

        return obj

    @staticmethod
    def _get_map_from_file(maps, prop):
        """Initialise the `.Map` from a `.~marvin.tools.maps.Maps` file."""

        if prop.channel is not None:
            channel_idx = prop.channel.idx
            value = maps.data[prop.name].data[channel_idx]
            ivar = maps.data[prop.name + '_ivar'].data[channel_idx] if prop.ivar else None
            mask = maps.data[prop.name + '_mask'].data[channel_idx] if prop.mask else None
        else:
            value = maps.data[prop.name].data
            ivar = maps.data[prop.name + '_ivar'].data if prop.ivar else None
            mask = maps.data[prop.name + '_mask'].data if prop.mask else None

        return value, ivar, mask

    @staticmethod
    def _get_map_from_db(maps, prop):
        """Initialise the `.Map` from the DB."""

        mdb = marvin.marvindb

        if not mdb.isdbconnected:
            raise marvin.core.exceptions.MarvinError('No db connected')

        if sqlalchemy is None:
            raise marvin.core.exceptions.MarvinError('sqlalchemy required to access the local DB.')

        assert prop.model is not None
        table = getattr(mdb.dapdb, prop.model)

        fullname_value = prop.db_column()
        props = [getattr(table, fullname_value)]
        if prop.ivar:
            fullname_ivar = prop.db_column(ext='ivar')
            props.append(getattr(table, fullname_ivar))
        if prop.mask:
            fullname_mask = prop.db_column(ext='mask')
            props.append(getattr(table, fullname_mask))

        n_props = len(props)

        ivar = None
        mask = None
        tmp = mdb.session.query(*props).filter(table.file_pk == maps.data.pk).use_cache(
            maps.cache_region).order_by(table.spaxel_index).all()

        shape = (int(np.sqrt(len(tmp))), int(np.sqrt(len(tmp))), n_props)
        newtmp = np.array(tmp).reshape(shape).T
        value = newtmp[0, :, :]
        if prop.ivar:
            ivar = newtmp[1, :, :]
        if prop.mask:
            mask = newtmp[2, :, :]
            mask = mask.astype(int)

        return value, ivar, mask

    @staticmethod
    def _get_map_from_api(maps, prop):
        """Initialise the `.Map` from the API."""

        url = marvin.config.urlmap['api']['getmap']['url']

        url_full = url.format(
            **{'name': maps.plateifu,
               'property_name': prop.name,
               'channel': prop.channel.name if prop.channel else None,
               'bintype': maps.bintype.name,
               'template': maps.template.name})

        try:
            response = marvin.api.api.Interaction(url_full,
                                                  params={'release': maps._release})
        except Exception as ee:
            raise marvin.core.exceptions.MarvinError(
                'found a problem when getting the map: {0}'.format(str(ee)))

        data = response.getData()

        if data is None:
            raise marvin.core.exceptions.MarvinError(
                'something went wrong. Error is: {0}'.format(response.results['error']))

        value = np.array(data['value'])
        ivar = np.array(data['ivar']) if data['ivar'] is not None else None
        mask = np.array(data['mask']) if data['mask'] is not None else None

        return value, ivar, mask

    def getSpaxel(self, **kwargs):
        """Returns a `~marvin.tools.spaxel.Spaxel`."""

        if not self._maps:
            raise ValueError('this Map does not have an associate parent Maps.')

        return self._maps.getSpaxel(**kwargs)

    def save(self, path, overwrite=False):
        """Pickle the map to a file.

        This method will fail if the map is associated to a Maps loaded
        from the db.

        Parameters:
            path (str):
                The path of the file to which the ``Map`` will be saved.
                Unlike for other Marvin Tools that derive from
                :class:`~marvin.tools.core.MarvinToolsClass`, ``path`` is
                mandatory for ``Map`` given that the there is no default
                path for a given map.
            overwrite (bool):
                If True, and the ``path`` already exists, overwrites it.
                Otherwise it will fail.

        Returns:
            path (str):
                The realpath to which the file has been saved.
        """
        # check for file extension
        if not os.path.splitext(path)[1]:
            path = os.path.join(path + '.mpf')

        return marvin.core.marvin_pickle.save(self, path=path, overwrite=overwrite)

    @classmethod
    def restore(cls, path, delete=False):
        """Restore a Map object from a pickled file.

        If ``delete=True``, the pickled file will be removed after it has been
        unplickled. Note that, for map objects instantiated from a Maps object
        with ``data_origin='file'``, the original file must exists and be
        in the same path as when the object was first created.
        """
        return marvin.core.marvin_pickle.restore(path, delete=delete)

    @property
    def masked(self):
        """Return a masked array using the recommended masks."""

        assert self.mask is not None, 'mask is None'

        default_params = self._datamodel.parent.get_default_plot_params()
        labels = default_params['default']['bitmasks']

        return np.ma.array(self.value, mask=self.pixmask.get_mask(labels, dtype=bool))

    @property
    def std(self):
        """The standard deviation of the measurement."""

        return self.error

    @staticmethod
    def _add_ivar(ivar1, ivar2, *args, **kwargs):
        return 1. / ((1. / ivar1 + 1. / ivar2))

    @staticmethod
    def _mul_ivar(ivar1, ivar2, value1, value2, value12):
        with np.errstate(divide='ignore', invalid='ignore'):
            sig1 = 1. / np.sqrt(ivar1)
            sig2 = 1. / np.sqrt(ivar2)
            sig12 = abs(value12) * ((sig1 / abs(value1)) + (sig2 / abs(value2)))
            ivar12 = 1. / sig12**2
        return ivar12

    @staticmethod
    def _pow_ivar(ivar, value, power):
        if ivar is None:
            return np.zeros(value.shape)
        else:
            sig = np.sqrt(1. / ivar)
            sig_out = value**power * power * sig * value
            ivar = 1 / sig_out**2.
            ivar[np.isnan(ivar) | np.isinf(ivar)] = 0
            return ivar

    @staticmethod
    def _log10_ivar(ivar, value):
        return np.log10(np.e) * ivar**-0.5 / value

    @staticmethod
    def _unit_propagation(unit1, unit2, op):
        ops = {'+': operator.add, '-': operator.sub, '*': operator.mul, '/': operator.truediv}

        if op in ['*', '/']:
            unit12 = ops[op](unit1, unit2)
        else:
            if unit1 == unit2:
                unit12 = unit1
            else:
                warnings.warn('Units do not match for map arithmetic.', UserWarning)
                unit12 = None

        return unit12

    def _arith(self, term2, op, reflexive=False):
        if isinstance(term2, (Map, EnhancedMap)):
            map_out = self._map_arith(term2, op)
        else:
            map_out = self._scalar_arith(term2, op, reflexive)
        return map_out

    def _scalar_arith(self, scalar, op, reflexive):
        """Do map arithmetic and correctly handle map attributes."""

        ops = {'+': operator.add, '-': operator.sub, '*': operator.mul, '/': operator.truediv}

        ivar_func = {'+': lambda ivar, c: ivar, '-': lambda ivar, c: ivar,
                     '*': lambda ivar, c: ivar / c**2, '/': lambda ivar, c: ivar * c**2}

        if not reflexive:
            args_value = (self.value, scalar)
        else:
            args_value = (scalar, self.value)
            ivar_func['/'] = lambda ivar, c: ivar

        with np.errstate(divide='ignore', invalid='ignore'):
            map_value = ops[op](*args_value)

        map1_ivar = self.ivar if self.ivar is not None else np.zeros(self.shape)
        map_ivar = ivar_func[op](map1_ivar, scalar)

        map_mask = self.mask if self.mask is not None else np.zeros(self.shape, dtype=int)

        bad = np.isnan(map_value) | np.isinf(map_value)
        map_mask[bad] = map_mask[bad] | self.pixmask.labels_to_value('DONOTUSE')

        return EnhancedMap(value=map_value, unit=self.unit, ivar=map_ivar, mask=map_mask,
                           datamodel=self._datamodel, pixmask_flag=self.pixmask_flag, copy=True)

    def _map_arith(self, map2, op):
        """Do map arithmetic and correctly handle map attributes."""

        ops = {'+': operator.add, '-': operator.sub, '*': operator.mul, '/': operator.truediv}

        assert self.shape == map2.shape, 'Cannot do map arithmetic on maps of different shapes.'

        ivar_func = {'+': self._add_ivar, '-': self._add_ivar,
                     '*': self._mul_ivar, '/': self._mul_ivar}

        with np.errstate(divide='ignore', invalid='ignore'):
            map12_value = ops[op](self.value, map2.value)

        map1_ivar = self.ivar if self.ivar is not None else np.zeros(self.shape)
        map2_ivar = map2.ivar if map2.ivar is not None else np.zeros(map2.shape)
        map12_ivar = ivar_func[op](map1_ivar, map2_ivar, self.value, map2.value, map12_value)
        map12_ivar[np.isnan(map12_ivar) | np.isinf(map12_ivar)] = 0

        map1_mask = self.mask if self.mask is not None else np.zeros(self.shape, dtype=int)
        map2_mask = map2.mask if map2.mask is not None else np.zeros(map2.shape, dtype=int)
        map12_mask = map1_mask | map2_mask

        bad = np.isnan(map12_value) | np.isinf(map12_value)
        map12_mask[bad] = map12_mask[bad] | self.pixmask.labels_to_value('DONOTUSE')

        map12_unit = self._unit_propagation(self.unit, map2.unit, op)

        return EnhancedMap(value=map12_value, unit=map12_unit, ivar=map12_ivar, mask=map12_mask,
                           datamodel=self._datamodel, pixmask_flag=self.pixmask_flag, copy=True)

    def __add__(self, map2):
        """Add two maps."""
        return self._arith(map2, '+')

    def __sub__(self, map2):
        """Subtract two maps."""
        return self._arith(map2, '-')

    def __mul__(self, map2):
        """Multiply two maps."""
        return self._arith(map2, '*')

    def __div__(self, map2):
        """Divide two maps."""
        return self._arith(map2, '/')

    def __truediv__(self, map2):
        """Divide two maps."""
        return self.__div__(map2)

    def __radd__(self, other):
        """Reflexive add."""
        return self._arith(other, '+', reflexive=True)

    def __rsub__(self, other):
        """Reflexive subtract."""
        return self._arith(other, '-', reflexive=True)

    def __rmul__(self, other):
        """Reflexive multiply."""
        return self._arith(other, '*', reflexive=True)

    def __rdiv__(self, other):
        """Reflexive divide."""
        return self._arith(other, '/', reflexive=True)

    def __rtruediv__(self, other):
        """Reflexive divide."""
        return self.__rdiv__(other)

    def __pow__(self, power):
        """Raise map to power.

        Parameters:
            power (float):
               Power to raise the map values.

        Returns:
            map (:class:`~marvin.tools.quantities.EnhancedMap` object)
        """
        value = self.value**power
        ivar = self._pow_ivar(self.ivar, self.value, power)
        unit = self.unit**power

        return EnhancedMap(value=value, unit=unit, ivar=ivar, mask=self.mask,
                           datamodel=self._datamodel, pixmask_flag=self.pixmask_flag, copy=True)

    def inst_sigma_correction(self):
        """Correct for instrumental broadening.

        Correct observed stellar or emission line velocity dispersion
        for instrumental broadening.
        """

        if self.datamodel.name == 'stellar_sigma':

            if 'MPL-4' in self._datamodel.parent.aliases:
                raise marvin.core.exceptions.MarvinError(
                    'Instrumental broadening correction not implemented for MPL-4.')
            elif 'MPL-6' in self._datamodel.parent.aliases:
                raise marvin.core.exceptions.MarvinError(
                    'The stellar sigma corrections in MPL-6 are unreliable. Please use MPL-7.')

            map_corr = self.getMaps()['stellar_sigmacorr']

        elif self.datamodel.name == 'emline_gsigma':
            map_corr = self.getMaps().getMap(property_name='emline_instsigma',
                                             channel=self.datamodel.channel.name)

        else:
            raise marvin.core.exceptions.MarvinError(
                'Cannot correct {0} for instrumental broadening.'.format(self.datamodel.full()))

        sigcorr = (self**2 - map_corr**2)**0.5
        sigcorr.ivar = (sigcorr.value / self.value) * self.ivar
        sigcorr.ivar[self.ivar == 0] = 0
        sigcorr.ivar[map_corr.value >= self.value] = 0
        sigcorr.value[map_corr.value >= self.value] = 0

        sigcorr._show_datamodel = True

        return sigcorr

    def specindex_correction(self):
        """Correct spectral index measurements for velocity dispersion.
        """
        if self.datamodel.name == 'specindex':

            if 'MPL-4' in self._datamodel.parent.aliases:
                raise marvin.core.exceptions.MarvinError(
                    'Velocity dispersion correction not implemented for MPL-4.')

            corr_name = '_'.join((self.datamodel.name, 'corr', self.datamodel.channel.name))
            map_corr = self.getMaps()[corr_name]

        else:
            raise marvin.core.exceptions.MarvinError(
                'Cannot correct {0} for velocity dispersion.'.format(self.datamodel.full()))

        if self.unit in ['Angstrom', '']:
            specindex_corr = self * map_corr

        elif self.unit == 'mag':
            specindex_corr = self + map_corr

        else:
            raise marvin.core.exceptions.MarvinError(
                'Specindex unit must be "Angstrom", "mag", or "" (i.e., dimensionless)')

        specindex_corr._show_datamodel = True

        return specindex_corr

    @add_doc(marvin.utils.plot.map.plot.__doc__)
    def plot(self, *args, **kwargs):
        return marvin.utils.plot.map.plot(dapmap=self, *args, **kwargs)


class EnhancedMap(Map):
    """Creates a Map that has been modified."""

    def __new__(cls, value, unit, *args, **kwargs):
        ignore = ['datamodel']
        [kwargs.pop(it) for it in ignore if it in kwargs]
        return super(EnhancedMap, cls).__new__(cls, value, unit=unit, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        self._datamodel = kwargs.get('datamodel', None)
        self._show_datamodel = False

    def __repr__(self):
        return ('<Marvin EnhancedMap>\n{0!r} {1!r}').format(self.value, self.unit.to_string())

    def __deepcopy__(self, memo):
        return EnhancedMap(value=deepcopy(self.value, memo), unit=deepcopy(self.unit, memo),
                           ivar=deepcopy(self.ivar, memo), mask=deepcopy(self.mask, memo),
                           pixmask_flag=deepcopy(self.pixmask_flag, memo), copy=True)

    def _init_map_from_maps(self):
        raise AttributeError("'EnhancedMap' has no attribute '_init_map_from_maps'.")

    def _get_from_file(self):
        raise AttributeError("'EnhancedMap' has no attribute '_get_from_file'.")

    def _get_from_db(self):
        raise AttributeError("'EnhancedMap' has no attribute '_get_from_db'.")

    def _get_from_api(self):
        raise AttributeError("'EnhancedMap' has no attribute '_get_from_api'.")

    def inst_sigma_correction(self):
        """Override Map.inst_sigma_correction with AttributeError."""
        raise AttributeError("'EnhancedMap' has no attribute 'inst_sigma_correction'.")

    @property
    def datamodel(self):
        if self._show_datamodel:
            return self._datamodel
        else:
            return None

    @datamodel.setter
    def datamodel(self, value):
        self._datamodel = value
        self._show_datamodel = True



#!/usr/bin/env python
# encoding: utf-8

'''
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-02-23 14:13:28 by Brian Cherinka
    2016-02-23 - made some test sample forms with wtforms+sqlalchemy - B. Cherinka
    2016-03-02 - generalized the form to build all forms - B. Cherinka
'''

from __future__ import division, print_function

import re
import six
import sys
import warnings
from collections import OrderedDict, defaultdict
from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property
from sqlalchemy.inspection import inspect as sa_inspect
from sqlalchemy.orm.attributes import InstrumentedAttribute
from wtforms import Field, SelectMultipleField, StringField, SubmitField, ValidationError, validators
from wtforms.widgets import TextInput
from wtforms_alchemy import model_form_factory
from marvin import config, marvindb
from marvin.core.exceptions import MarvinUserWarning
from marvin.utils.general.structs import FuzzyDict

# flask_wtf does not work locally - OUTSIDE APPLICATON CONTEXT;
# need some kind of toggle for web version and not???
if config._inapp:
    from flask_wtf import Form
else:
    from wtforms import Form

__all__ = ['MarvinForm']


def tree():
    return defaultdict(tree)


# Base form class
BaseModelForm = model_form_factory(Form)


class ModelForm(BaseModelForm):
    ''' sub class a new WTF ModelForm so it works with Flask-WTF in APP mode;
    for auto CSRF tokens...who knows...
    '''
    @classmethod
    def get_session(self):
        return marvindb.session

# Builds a dictionary for modelclasses with key ClassName and value SQLalchemy model class
# modelclasses = marvindb.buildUberClassDict()


# Class factory
def formClassFactory(name, model, baseclass):
    ''' Generates a new WTForm Class based on SQLalchemy Model Class.

    Subclasses a base WTF Form class that also contains the SQLAlchemy
    Model Class information inside it.

    Each class contains as attributes:
        Meta = a class called Meta.  Meta.model contains the SQLalchemy ModelClass
        data = a dictionary of parameters: form input that gets mapped to the sqlalchemy parameter
        errors = a dictionary of errors returned by invalid form validation
        validate = a method to validate all elements in this form
        parameter_X = a WTForm Field mapped to respective sqlalchemy table column

        e.g.
        The ModelClass IFUDesign mapped to mangadatadb.ifu_design sql table gets transformed into
        WTForm IFUDesignForm, with IFUDesignForm.Meta.model = marvin.db.models.DataModelClasses.IFUDesign

    Parameters:
        name (str):
            The name of the Form Class
        mdoel (class):
            The SQLAlchemy Model Class
        baseclass (class):
            The base class to sub class from

    Returns:
        the new WTF form subclass
    '''

    Meta = type('Meta', (object,), {'model': model})
    newclass = type(name, (baseclass,), {'Meta': Meta})
    return newclass

# build a wtform select field for operators; tested but no longer used
# can't seem to attach operator field to every individual parameter
opdict = {'le': '<=', 'ge': '>=', 'gt': '>', 'lt': '<', 'ne': '!=', 'eq': '='}
ops = [(key, val) for key, val in opdict.items()]
# operator = SelectField(u'Operator', choices=ops)


class ParamFxnLookupDict(dict):
    ''' Parameter function lookup for new function expressions
    '''

    def __getitem__(self, key):

        lowkey = key.lower()
        mykeys = list(self.keys())

        inkey = lowkey in mykeys
        if not inkey:
            raise KeyError('{0} does not match any column.'.format(lowkey))
        else:
            keycount = mykeys.count(lowkey)
            if keycount > 1:
                raise KeyError('{0} matches multiple parameters in the lookup'
                               ' table'.format(lowkey))
            else:
                return dict.__getitem__(self, lowkey)


class ParamFormLookupDict(dict):

    def __init__(self, **kwargs):
        self.allspaxels = kwargs.get('allspaxels', None)
        self._release = kwargs.get('release', config.release)
        self._init_shortcuts()

    def __contains__(self, value):
        ''' Override the contains '''

        try:
            key = self[value]
        except KeyError as e:
            key = None

        return key is not None

    def __getitem__(self, key):
        """Checks if `key` is a unique column name and return the value."""

        # Init the shortcuts
        self._init_shortcuts()

        # Applies shortcuts
        keySplits = self._apply_shortcuts(key)

        # Get key matches
        matches = self._get_matches(keySplits)

        # Check DAP Junk keys
        matches = self._check_for_junk(matches)

        # Return matched key
        return dict.__getitem__(self, self._get_good_match(key, matches))

    def _check_for_junk(self, matches):
        ''' check for Junk matches and return the correct key '''
        isdapjunk = any(['mangadapdb.spaxelprop' in m for m in matches])
        if isdapjunk:
            junkmatches = [m for m in matches if (self.allspaxels and 'clean'
                           not in m) or (not self.allspaxels and 'clean' in m)]
            keySplits = self._apply_shortcuts(junkmatches[0])
            matches = self._get_matches(keySplits)
        return matches

    def _get_good_match(self, key, matches):
        ''' Check if the key match is good '''
        if len(matches) == 0:
            # No matches. This returns the standards KeyError from dict
            raise KeyError('{0} does not match any column.'.format(key))
        elif len(matches) == 1:
            # One match: returns True
            return matches[0]
        else:
            # Multiple results. Raises a custom error.
            raise KeyError(
                '{0} matches multiple parameters in the lookup table. Try one of: {1}'
                .format(key, ', '.join(matches)))

    def mapToColumn(self, keys):
        """Returns the model class column in the WFTForm."""

        if not isinstance(keys, (list, tuple)):
            if isinstance(keys, six.string_types):
                keys = [keys]
            else:
                keys = list(keys)

        columns = []
        for key in keys:
            keySplits = self._apply_shortcuts(key)
            matches = self._get_matches(keySplits)
            matches = self._check_for_junk(matches)
            key = self._get_good_match(key, matches)
            wtfForm = self[key]
            column = key.split('.')[-1]
            columns.append(getattr(wtfForm.Meta.model, column))

        if len(columns) == 1:
            return columns[0]
        else:
            return columns

    def _init_table_shortcuts(self):
        ''' initialize the table shortcuts '''

        self._schemaShortcuts = {'datadb': 'mangadatadb', 'dapdb': 'mangadapdb', 'sampledb': 'mangasampledb',
                                 'auxdb': 'mangaauxdb'}

        self._tableShortcuts = {'ifu': 'ifudesign', 'cube_header_keyword': 'fits_header_keyword',
                                'cube_header_value': 'fits_header_value', 'maps_header_keyword': 'header_keyword',
                                'maps_header_value': 'header_value'}
        self._set_junk_shortcuts()

    def _init_name_shortcuts(self):
        ''' initialize the name shortcuts

        e.g {'haflux': 'emline_gflux_ha_6564',
             'g_r': 'elpetro_mag_g_r',
             'abs_g_r': 'elpetro_absmag_g_r'}

        '''
        from marvin.utils.datamodel.query.base import query_params
        short = query_params.list_params(name_type='short')
        name = query_params.list_params(name_type='name')
        self._nameShortcuts = OrderedDict(zip(short, name))

    def _init_full_shortcuts(self):
        ''' initialize any full name shortcuts '''
        self._fullShortcuts = {'cube_header.keyword':'fits_header_keyword.label',
                               'cube_header.value':'fits_header_value.value',
                               'maps_header.keyword':'header_keyword.name',
                               'maps_header.value':'header_value.value'}

    def _init_shortcuts(self):
        ''' initialize the shortcuts '''
        self._init_full_shortcuts()
        self._init_table_shortcuts()
        self._init_name_shortcuts()

    def _apply_shortcuts(self, key, reverse=None):
        ''' Apply the shortcuts to the key '''

        full = self._reverse_shorts(self._fullShortcuts) if reverse else self._fullShortcuts
        schema = self._reverse_shorts(self._schemaShortcuts) if reverse else self._schemaShortcuts
        table = self._reverse_shorts(self._tableShortcuts) if reverse else self._tableShortcuts
        name = self._reverse_shorts(self._nameShortcuts) if reverse else self._nameShortcuts

        # Apply any full name shortcuts
        if key in full:
            return full[key].split('.')

        # Gets the paths that match the key
        keySplits = key.split('.')

        # Apply schema shortcuts
        if len(keySplits) >= 3 and keySplits[-3] in schema:
            keySplits[-3] = schema[keySplits[-3]]

        # Applies table shortcuts
        if len(keySplits) >= 2 and keySplits[-2] in table:
            keySplits[-2] = table[keySplits[-2]]

        # Applies name shortcuts
        if len(keySplits) >= 1 and keySplits[-1] in name:
            keySplits[-1] = name[keySplits[-1]]

        return keySplits

    def get_real_name(self, key):
        ''' Returns the real full key given some shortcut names '''
        keySplits = self._apply_shortcuts(key)
        return '.'.join(keySplits)

    def get_shortcut_name(self, key):
        ''' Returns the shortcutted full name given a real key '''

        keySplits = self._apply_shortcuts(key, reverse=True)
        return '.'.join(keySplits)

    def _reverse_shorts(self, data):
        ''' Reverse a shorts dictionary '''
        return {v: re.sub(r"spaxelprop[1-9]", "spaxelprop", k) for k, v in data.items()}

    def _set_junk_shortcuts(self):
        ''' Sets the DAP spaxelprop shortcuts based on MPL '''

        spaxname = 'spaxelprop' if self.allspaxels else 'cleanspaxelprop'

        # available spaxel property tables
        dapcut = {}
        spdict = marvindb.spaxelpropdict
        if spdict:
            # create a shortcut for each table that points to the current release
            current = spdict.get(self._release, None)
            for release, name in spdict.items():
                if current:
                    cut = {name.lower(): current.lower().replace('spaxelprop', spaxname)}
                    dapcut.update(cut)

            # since for DRs spaxelprop may not be in the dictionary a la MPL4: SpaxelProp
            # we need to add it in so the shortcut "spaxelprop" can correctly point to the right db param
            if 'spaxelprop' not in dapcut and current:
                dapcut.update({'spaxelprop': current.lower().replace('spaxelprop', spaxname)})

        # update the main dictionary
        self._tableShortcuts.update(dapcut)

    def _get_matches(self, keySplits):
        ''' Get the matches from a set of key splits '''
        matches = [path for path in self
                   if all([keySplits[-1 - ii] == path.split('.')[-1 - ii]
                           for ii in range(len(keySplits))])]
        return matches


# Custom [Better]TagList Field for Search Parameter Box
class TagListField(Field):
    widget = TextInput()

    def _value(self):
        if self.data:
            return u', '.join(self.data)
        else:
            return u''

    def process_formdata(self, valuelist):
        if valuelist:
            self.data = [x.strip() for x in valuelist[0].split(',') if x]
        else:
            self.data = []


class BetterTagListField(TagListField):
    def __init__(self, label='', validators=None, remove_duplicates=True, **kwargs):
        super(BetterTagListField, self).__init__(label, validators, **kwargs)
        self.remove_duplicates = remove_duplicates

    def process_formdata(self, valuelist):
        super(BetterTagListField, self).process_formdata(valuelist)
        if self.remove_duplicates:
            self.data = list(self._remove_duplicates(self.data))

    @classmethod
    def _remove_duplicates(cls, seq):
        """Remove duplicates in a case insensitive, but case preserving manner"""
        dups = {}
        for item in seq:
            if item.lower() not in dups:
                dups[item.lower()] = True
                yield item


# Custom validator for MainForm
class ValidOperand(object):
    def __init__(self, opstring='[<>=]', message=None):
        self.opstring = opstring
        if not message:
            message = u'Field must contain at least a valid operand of {0}.'.format(self.opstring)
        self.message = message

    def __call__(self, form, field):
        infield = re.search(self.opstring, field.data)
        if not infield:
            raise ValidationError(self.message)


class SearchForm(Form):
    ''' Main Search Level WTForm for Marvin '''

    url = ("<a target='_blank' data-toggle='tooltip' title='Click to read more about Queries' "
           "href='https://sdss-marvin.readthedocs.io/en/stable/query/query/query_using.html'>Input Search Filter</a>")
    searchbox = StringField(url,
                            [validators.Length(min=3, message='Input must have at least 3 characters'),
                             validators.DataRequired(message='Input filter string required'),
                             ValidOperand('[<>=betweenradial]', message='Input must contain a valid operand.')])
    returnparams = SelectMultipleField("<a target='_blank'>Return Parameters</a>")
    submitsearch = SubmitField('Search')


class MarvinForm(object):
    ''' Core Marvin Form object. '''

    def __init__(self, *args, **kwargs):
        ''' Initializes a Marvin Form

        Generates all the WTForms from the SQLAlchemy ModelClasses defined in the MaNGA DB.

        _param_form_lookup = dictionary of all modelclass parameters
        of form {'SQLalchemy ModelClass parameter name': WTForm Class}
        '''

        self._release = kwargs.get('release', config.release)
        self.verbose = kwargs.get('verbose', False)
        if marvindb:
            self._modelclasses = FuzzyDict(marvindb.buildUberClassDict(release=self._release))
            self._param_form_lookup = ParamFormLookupDict(**kwargs)
            self._param_fxn_lookup = ParamFxnLookupDict()
            self._paramtree = tree()
            self._generateFormClasses(self._modelclasses)
            self._generateFxns()
            self.SearchForm = SearchForm
            self._cleanParams(**kwargs)

    def __repr__(self):
        nforms = len([f for f in self.__dict__.keys() if 'Form' in f])
        return ('<MarvinForm (release={0._release}, n_parameters={1}, n_functions={2}, '
                'n_forms={3})>'.format(self, len(self._param_form_lookup), len(self._param_fxn_lookup), nforms))

    def _generateFormClasses(self, classes):
        ''' Loops over all ModelClasses and generates a new WTForm class.  New form classes are named as [ModelClassName]Form.
            Sets the new form as an attribute on MarvinForm.  Also populates the _param_to_form_lookup dictonary with
            all ModelClass/WTForm parameters and their corresponding forms.

            e.g.  _param_form_lookup['name'] = marvin.tools.query.forms.IFUDesignForm
        '''

        for key, val in classes.items():
            classname = '{0}Form'.format(key)
            try:
                newclass = formClassFactory(classname, val, ModelForm)
            except Exception as e:
                if self.verbose:
                    warnings.warn('class {0} not Formable'.format(key), MarvinUserWarning)
            else:
                self.__setattr__(classname, newclass)
                self._loadParams(newclass)

    def _loadParams(self, newclass):
        ''' Loads all parameters from wtforms into a dictionary with
            key, value = {'parameter_name': 'parent WTForm name'}.
            Ignores hidden attributes and the Meta class
        '''

        model = newclass.Meta.model
        schema = model.__table__.schema
        tablename = model.__table__.name

        mapper = sa_inspect(model)
        for key, item in mapper.all_orm_descriptors.items():
            if isinstance(item, (hybrid_property, hybrid_method)):
                key = key
            elif isinstance(item, InstrumentedAttribute):
                key = item.key
            else:
                continue

            lookupKeyName = schema + '.' + tablename + '.' + key
            self._param_form_lookup[lookupKeyName] = newclass
            self._paramtree[newclass.Meta.model.__name__][key]

    def callInstance(self, form, params=None, **kwargs):
        ''' Creates an instance of a specified WTForm.  '''
        return form(**params) if params else form(**kwargs)

    def _generateFxns(self):
        ''' Generate the fxn dictionary

            TODO: make this general and not hard-coded

            The key is the function name used in the query syntax.
            The value is the method call that lives in Query

        '''
        self._param_fxn_lookup['npergood'] = '_get_percent'
        self._param_fxn_lookup['radial'] = '_radial_query'

    def _getDapKeys(self):
        ''' Returns the DAP keys from the Junk tables
        '''
        dapkeys = [k for k in self._param_form_lookup.keys() if 'mangadapdb.spaxelprop' in k]
        dapkeys.sort()
        return dapkeys

    def _cleanParams(self, **kwargs):
        ''' Clean up the parameter-form lookup dictionary '''

        # remove keys for pk, mangadatadb.sample, test_, and cube_header
        new = ParamFormLookupDict(**kwargs)
        for k, v in self._param_form_lookup.items():
            if 'pk' not in k and \
               'mangadatadb.sample' not in k and \
               'test_' not in k and \
               'cube_header' not in k:
                new[k] = v

        # make new dictionary
        self._param_form_lookup = new

    def look_up_table(self, table):
        ''' Look up a database table and return the ModelClass '''

        try:
            table_class = self._modelclasses[table]
        except (ValueError, TypeError) as e:
            table_class = None

        return table_class



#!/usr/bin/env python
# encoding: utf-8
"""

general.py

Licensed under a 3-clause BSD license.

Revision history:
    17 Feb 2016 J. Sánchez-Gallego
      Initial version
    18 Feb 2016 B. Cherinka
        Added buildRouteMap API call

"""

from __future__ import division
from __future__ import print_function

from flask import jsonify, Response, request
from flask_classful import route

from brain.api.base import processRequest
from brain.api.general import BrainGeneralRequestsView
from brain.utils.general import validate_user, get_db_user
from brain.utils.general.decorators import public
from marvin import marvindb, config
from marvin.utils.general import mangaid2plateifu as mangaid2plateifu
from marvin.utils.general import get_nsa_data
from marvin.api.base import arg_validate as av
from flask_jwt_extended import create_access_token
import json


def get_drpver(release=None):
    ''' Get the drpver from the input release from the request '''
    if not release:
        release = config.release

    drpver, dapver = config.lookUpVersions(release)
    return drpver


class GeneralRequestsView(BrainGeneralRequestsView):

    @route('/mangaid2plateifu/<mangaid>/', endpoint='mangaid2plateifu', methods=['GET', 'POST'])
    @av.check_args()
    def mangaid2plateifu(self, args, mangaid):
        """ Returns a plateifu given a mangaid

        .. :quickref: General; Returns a plateifu given a mangaid

        :param mangaid: The name of the cube as mangaid
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string plateifu: id of cube as plateifu
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/general/mangaid2plateifu/1-209232/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {8485-1901"}
           }

        """
        # get drpver
        drpver = get_drpver(release=args.get('release', None))

        try:
            plateifu = mangaid2plateifu(mangaid, mode='db', drpver=drpver)
            self.results['data'] = plateifu
            self.results['status'] = 1
        except Exception as ee:
            self.results['status'] = -1
            self.results['error'] = ('manga2plateifu failed with error: {0}'.format(str(ee)))

        return jsonify(self.results)

    @route('/nsa/full/<mangaid>/', endpoint='nsa_full', methods=['GET', 'POST'])
    @av.check_args()
    def get_nsa_data(self, args, mangaid):
        """Returns the NSA data for a given mangaid from the full catalogue.

        .. :quickref: General; Returns for a given mangaid the NSA data from the full catalog

        :param mangaid: The name of the cube as mangaid
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json dict nsa_data: dict of all the NSA parameters
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/general/nsa/full/1-209232/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"aid": 0,
                       "asymmetry":[-0.0217, 0.00517, -0.0187, ...],
                       "bastokes": [[0.7290, 0.9677, ...], [], ...],
                       ...
                      }
           }

        """
        # get drpver
        drpver = get_drpver(release=args.get('release', None))

        try:
            nsa_data = get_nsa_data(mangaid, mode='local', source='nsa', drpver=drpver)
            self.results['data'] = nsa_data
            self.results['status'] = 1
        except Exception as ee:
            self.results['status'] = -1
            self.results['error'] = 'get_nsa_data failed with error: {0}'.format(str(ee))

        # these should be jsonify but switching back to json.dumps until fucking Utah gets with the fucking picture
        return Response(json.dumps(self.results), mimetype='application/json')

    @route('/nsa/drpall/<mangaid>/', endpoint='nsa_drpall', methods=['GET', 'POST'])
    @av.check_args()
    def get_nsa_drpall_data(self, args, mangaid):
        """Returns the NSA data in drpall for a given mangaid.

        .. :quickref: General; Returns for a given mangaid the NSA data from the drpall file

        Note that this always uses the drpver/drpall versions that are default in the server.

        :param mangaid: The name of the cube as mangaid
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json dict nsa_data: dict of the NSA parameters from the DRPall file
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/general/nsa/drpall/1-209232/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"camcol": 2,
                       "elpetro_absmag": [-14.3422, -15.7994,-17.0133, ...],
                       "elpetro_amivar": [],
                       ...
                      }
           }

        """
        # get drpver
        drpver = get_drpver(release=args.get('release', None))

        try:
            nsa_data = get_nsa_data(mangaid, mode='local', source='drpall', drpver=drpver)
            self.results['data'] = nsa_data
            self.results['status'] = 1
        except Exception as ee:
            self.results['status'] = -1
            self.results['error'] = 'get_nsa_data failed with error: {0}'.format(str(ee))

        return Response(json.dumps(self.results), mimetype='application/json')

    @public
    @route('/login/', methods=['POST'], endpoint='login')
    def login(self):
        ''' Server-Side login to generate a new token '''

        result = {}

        # check the form
        form = processRequest(request=request)
        if form is None:
            result['error'] = 'Request has no form data!'
            return jsonify(result), 400

        # get username and password
        username = form.get('username', None)
        password = form.get('password', None)
        # return if no valid login form data
        if not username or not password:
            result['error'] = 'Missing username and/or password!'
            return jsonify(result), 400
        username = username.strip()
        password = password.strip()

        # validate the user with htpassfile or trac username
        is_valid, user, result = validate_user(username, password, request=request)

        # User code goes here
        if is_valid:
            user = get_db_user(username, password, dbsession=marvindb.session, user_model=marvindb.datadb.User, request=request)
            if user and user.check_password(password):
                # generate token if valid
                access_token = create_access_token(identity=user.username, fresh=True)
                return jsonify(access_token=access_token), 200
        else:
            msg = result['error'] if 'error' in result else ''
            result['error'] = 'Not valid login. Bad username or password. {0}'.format(msg)
            return jsonify(result), 401



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: vacs.py
# Project: tools
# Author: Brian Cherinka
# Created: Sunday, 8th September 2019 11:16:35 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Tuesday, 3rd December 2019 2:43:32 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import contextlib
import inspect
import six
import os
import importlib
import types
import numpy as np
from marvin.contrib.vacs.base import VACContainer, VACMixIn
from marvin import config, log
from marvin.core.exceptions import MarvinError
from marvin.utils.general import parseIdentifier, get_drpall_table
from decorator import decorator

from astropy.io import fits
from astropy.table import Table

import sdss_access.sync


def check_release(*args, **kwargs):
    ''' Decorator to check VAC release against Marvin config release  '''
    @decorator
    def decorated_function(ff, *args, **kwargs):
        inst = args[0]
        if inst.release != config.release:
            raise AssertionError('Mismatch between VAC release and current Marvin release.'
                                 'Please reinstantiate your VACs.')

        return ff(*args, **kwargs)
    return decorated_function


class check_vac_release(object):
    ''' Decorate all functions in a class to run only if VAC release matches Marvin release '''

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, decorated_class):
        for attr in inspect.getmembers(decorated_class, inspect.isfunction):
            # only decorate public functions
            if attr[0][0] != '_':
                setattr(decorated_class, attr[0],
                        check_release(*self.args, **self.kwargs)(getattr(decorated_class, attr[0])))
        return decorated_class


@check_vac_release()
class VACs(VACContainer):
    ''' A class to interface with the MaNGA VACs

    A container class to interact with all MaNGA VACs that
    have been contributed into Marvin.  Currently, this container
    only works with local VACs.  Remote-only access to VACs currently
    does not exist.

    Attributes:
        release (str):
            The current MPL/DR version of the loaded VACs
        vacname (object):
            A specific VAC object.  "vacname" is the name of the VAC, e.g. galaxyzoo

    Example:
        >>> from marvin.tools.vacs import VACs
        >>> vacs = VACs()
    '''

    @classmethod
    def _reset_vacs(cls):
        ''' reset the VAC class to a blank slate '''
        if hasattr(cls, '_vacs'):
            for vac in cls._vacs:
                # close the HDU file resource
                vv = getattr(cls, vac)
                if hasattr(vv, '_data') and getattr(vv, '_data') is not None:
                    vv._data.close()
                # delete the attribute
                delattr(cls, vac)

    def __new__(cls, *args, **kwargs):
        ''' load relevant VACMixIns upon new class creation '''
        # reset the vacs
        cls._reset_vacs()
        # initiate vacs for a given release
        cls._vacs = []
        cls.release = config.release
        for subvac in VACMixIn.__subclasses__():
            # Exclude any hidden VACs
            if subvac._hidden and (not subvac._hidden_for or 
                                   (subvac._hidden_for == cls.release)):
                continue

            # Excludes VACs from other releases
            if config.release not in subvac.version:
                continue

            sv = subvac()
            if sv.summary_file is None:
                log.info('VAC {0} has no summary file to load.  Skipping.'.format(sv.name))
                continue

            # create the VAC data class
            if isinstance(sv.summary_file, list):
                # return a list of data classes if multiple summary files are provided
                if sv.data_container and type(sv.data_container) is dict:
                    # use any data container if provided
                    vacdc = {k: VACDataClass(subvac.name, subvac.description, v, 
                                             display_name=subvac.display_name, url=subvac.url) for k, v in sv.data_container.items()}
                else:
                    vacdc = [VACDataClass(subvac.name, subvac.description, s, 
                                          display_name=subvac.display_name, url=subvac.url) for s in sv.summary_file]
                    if len(vacdc) == 1:
                        vacdc = vacdc[0]
            elif isinstance(sv.summary_file, str):
                vacdc = VACDataClass(subvac.name, subvac.description, sv.summary_file, display_name=subvac.display_name, url=subvac.url)

            # add any custom methods to the VAC
            if hasattr(subvac, 'add_methods'):
                for method in getattr(subvac, 'add_methods'):
                    assert isinstance(method, six.string_types), 'method name must be a string'
                    svmod = importlib.import_module(subvac.__module__)
                    assert hasattr(svmod, method), 'method name must be defined in VAC module file'
                    setattr(vacdc, method, types.MethodType(
                        getattr(svmod, method), vacdc))

            # add vac to the container class
            cls._vacs.append(subvac.name)
            setattr(cls, subvac.name, vacdc)

        return super(VACs, cls).__new__(cls, *args, **kwargs)

    def __repr__(self):
        ''' repr for VAC container '''
        n_vacs = len(self._vacs)
        if n_vacs <= 7:
            return '<VACs ({0})>'.format(', '.join(self._vacs))
        return '<VACs (n_vacs={0})>'.format(n_vacs)

    def list_vacs(self):
        ''' list the available VACS '''
        return self._vacs

    def check_target(self, target, lomem=None):
        ''' check which VACS a MaNGA target is available in

        Checks for the target in each VAC and returns a
        dictionary of VAC names and a boolean indicating
        if the target is in that VAC

        Parameters:
            target (str):
                The name of the target
            lomem (str):
                If True, loads the data via context manager

        Returns:
            A dict of booleans
        '''
        vacs = {}
        for vac in self._vacs:
            if isinstance(self[vac], VACDataClass):
                vacs[vac] = self[vac].has_target(target, lomem=lomem)
            elif isinstance(self[vac], dict):
                vacs[vac] = {k: v.has_target(target, lomem=lomem) for k, v in self[vac].items()}
            elif isinstance(self[vac], list):
                vacs[vac] = [v.has_target(target, lomem=lomem) for v in self[vac]]
        return vacs
    
    def match_targets(self, targets: list = None, to_df: bool = True):
        """ Match a list of mangaid targets across all VACs

        Check a list of mangaid targets against all VACs and return a Pandas
        DataFrame (default) or a dictionary of booleans indicating which mangaids
        are in which VACs.  If no targets are provided, all targets from the 
        DRPall file of the currently set release will be checked.

        Parameters
        ----------
        targets : list, optional
            A list of mangaids, by default None
        to_df : bool, optional
            If True converts the dictionary to a pandas dataframe, by default True

        Returns
        -------
        pandas.DataFrame or dict
            Booleans indicating which mangaids are in which VACs

        Raises
        ------
        ValueError
            when targets is not a list
        ValueError
            when the targets are not a mangaid
        """

        if not targets:
            drpver, __ = config.lookUpVersions(self.release)
            targets = get_drpall_table(drpver)['mangaid'].tolist()

        if not isinstance(targets, list):
            raise ValueError('targets must be a list')

        # check target
        targtype = parseIdentifier(targets[0])
        if targtype != 'mangaid':
            raise ValueError('targets must be mangaids')

        vtargs = {"release": [self.release] * len(targets), "mangaid": targets}
        for vac in self._vacs:
            if isinstance(self[vac], VACDataClass):
                mm = self[vac].get_mangaids()
                vtargs[vac] = np.isin(targets, mm)
            elif isinstance(self[vac], dict):
                for kk, vv in self[vac].items():
                    vtargs[f'{vac}_{kk}'] = np.isin(targets, vv.get_mangaids())
            elif isinstance(self[vac], list):
                for ii, ee in enumerate(self[vac]):
                    vtargs[f'{vac}_{ii}'] = np.isin(targets, ee.get_mangaids())
        
        # by default convert to a pandas dataframe
        if to_df:
            import pandas as pd
            return pd.DataFrame.from_dict(vtargs)
        return vtargs

class VACDataClass(object):
    ''' A data class for a given VAC

    Parameters:
        name (str):
            The name of the VAC
        description (str):
            The description of the VAC
        path (str):
            The full path to the VAC summary file
        display_name (str):
            The display name of the VAC
        url (str):
            The URL to the VAC description page

    Attributes:
        name (str):
            The name of the VAC
        description (str):
            The description of the VAC
        data (HDUList):
            The Astropy FITS HDUList
        display_name (str):
            The display name of the VAC
        url (str):
            The URL to the VAC description page

    Example:
        >>> from marvin.tools.vacs import VACs
        >>> vacs = VACs()
        >>> vacs.galaxyzoo.data
        >>> vacs.galaxyzoo.get_table(ext=1)
    '''
    def __init__(self, name, description, path, display_name=None, url=None):
        self.name = name
        self.description = description
        self.display_name = display_name
        self.url = url
        self._path = path
        self._data = None

        if not sdss_access.sync.Access:
            raise MarvinError('sdss_access is not installed')
        else:
            self._rsync = sdss_access.sync.Access(release=config.release)

    def __repr__(self):
        name = self.name.title().replace('_', '')
        if not self._data:
            end = ''
        end = '' if not self._data else ', n_hdus={0}'.format(len(self.data))
        return '<{0}Data(description={1}{2})>'.format(name, self.description, end)

    @property
    def data(self):
        ''' The VAC FITS data '''
        if not self._data:
            self._check_file()
            self._data = fits.open(self._path)
        return self._data
    
    @contextlib.contextmanager
    def get_data(self, ext=1):
        ''' Get the data from a specific extension
        
        Context manager to return the data from a given extension.
        Useful for large FITS files that you don't want to load all at once.

        '''
        self._check_file()
        with fits.open(self._path) as hdulist:
            yield hdulist[ext].data
        del hdulist[ext].data
    
    def __del__(self):
        ''' Close the file on delete '''
        if self._data and isinstance(self._data, fits.HDUList):
            self._data.close()

    def info(self):
        ''' print the VAC HDU info '''
        self.data.info()

    def get_table(self, ext=None):
        ''' Create an Astropy table for a data extension

        Parameters:
            ext (int|str):
                The HDU extension name or number

        Returns:
            An Astropy table for the given extension
        '''
        if not ext:
            log.info('No HDU extension specified.  Defaulting to ext=1')
            ext = 1

        # check if extension is an image
        if self.data[ext].is_image:
            log.info('Ext={0} is not a table extension.  Cannot read.'.format(ext))
            return

        return Table.read(self._path, ext, format='fits')

    def download_vac(self):
        ''' Download a VAC to the local system '''
        if self._data:
            log.info('File already downloaded.')
            return

        log.info('Downloading VAC data..')
        self._rsync.reset()
        self._rsync.remote()
        self._rsync.add('', full=self._path)
        self._rsync.set_stream()
        self._rsync.commit()

    def _check_file(self):
        ''' check for a file locally, else downloads it '''

        # ensure file is of FITS format
        ext = os.path.splitext(self._path)[-1]
        assert '.fits' in ext, 'VAC summary file must be of FITS format'

        # download file
        if not self._rsync.exists('', full=self._path):
            self.download_vac()

    def has_target(self, target: str, lomem: bool = False) -> bool:
        ''' Checks if a target is contained within the VAC

        Checks if a target mangaid or plateifu is within the 
        first data extension of the VAC.  By default, uses the
        full data loaded into memory.  Use the lomem flag to
        instead load the data using a context manager.  This is useful
        for large FITS files in a low-memory environment, such as a web
        server.

        Parameters:
            target (str):
                The name of the target
            lomem (str):
                If True, loads the data via context manager

        Returns:
            A boolean indicating if the target is in the VAC or not
        '''
        assert isinstance(target, six.string_types), 'target must be a string'
        targ_type = parseIdentifier(target)
        ttypes = ['plateifu', 'mangaid']
        other = ttypes[ttypes.index(targ_type)-1]
        
        if not lomem:
            data = self.data[1].data
            assert targ_type in map(str.lower, data.columns.names), \
                'Identifier "{0}" is not available.  Try a "{1}".'.format(targ_type, other)
            return target in data[targ_type]

        with self.get_data(ext=1) as data:
            assert targ_type in map(str.lower, data.columns.names), \
                'Identifier "{0}" is not available.  Try a "{1}".'.format(targ_type, other)

            return target in data[targ_type]
        
    def get_mangaids(self, ext=1):
        """ Get the mangaids from the VAC """
        return self.data[ext].data['mangaid'].strip().tolist()
        



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: caching.py
# Project: db
# Author: Brian Cherinka
# Created: Friday, 24th January 2020 2:47:54 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Monday, 27th January 2020 6:36:07 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
from marvin import config
from hashlib import md5
from dogpile.cache.region import make_region
import os
import copy

try:
    import redis
except ImportError:
    redis = None

# DOGPILE CACHING SETUP

# dogpile cache regions.  A home base for cache configurations.
regions = {}

# create the dogpile directory for file-based caches
dogpath = os.environ.get('MANGA_SCRATCH_DIR', None)
if dogpath and os.path.isdir(dogpath):
    dogpath = dogpath
else:
    dogpath = os.path.expanduser('~')

dogroot = os.path.join(dogpath, 'dogpile_data')
if not os.path.isdir(dogroot):
    os.makedirs(dogroot)


# db hash key
def md5_key_mangler(key):
    """Receive cache keys as long concatenated strings;
    distill them into an md5 hash.

    """
    return md5(key.encode('ascii')).hexdigest()


def redis_key_mangler(key):
    ''' Set a redis key mangler

    Prefix the cache key for redis caches to distinguish dogpile
    caches in redis from other caches in redis.  See
    https://dogpilecache.sqlalchemy.org/en/latest/recipes.html#prefixing-all-keys-in-redis

    '''
    return "marvin:dogpile:" + key


# cache backend type
cache_type = {
    'null': {'name': 'dogpile.cache.null', 'args': {}},
    'file': {'name': 'dogpile.cache.dbm', 'args': {'filename': dogroot}},
    'redis': {'name': 'dogpile.cache.redis',
              'args': {'url': os.environ.get('SESSION_REDIS'),
                       'distributed_lock': True, 'lock_sleep': 1,
                       'lock_timeout': 5, 'thread_local_lock': False}}
}


def make_new_region(name='cache.dbm', backend='file', expiration=3600, redis_exp_multiplier=2):
    ''' make a new dogpile cache region

    Creates a new dogpile cache region.  Default is to create a file-based
    cache called cache.dbm, with an expiration of 1 hour.

    Parameters:
        name (str):
            The name of the file-based cache
        backend (str):
            The type of cache backend to use.  Default is file.
        expiration (int):
            Expiration time in seconds of dogpile cache.  Default is 3600 seconds.
        redis_exp_multiplier (int):
            Integer factor to compute redis cache expiration time from dogpile expiration.
            Default is 2x the expiration.
    '''
    # select the backend
    cache_backend = copy.deepcopy(cache_type[backend])

    # set the key mangler
    key_mangler = redis_key_mangler if backend == 'redis' else md5_key_mangler

    # create the cache filename
    if backend == 'file':
        fileroot = cache_backend['args'].get('filename', dogroot)
        cache_backend['args']['filename'] = os.path.join(fileroot, name)

    # update the redis expiration time
    if backend == 'redis':
        cache_backend['args']['redis_expiration_time'] = expiration * redis_exp_multiplier
        assert cache_backend['args']['url'] and cache_backend['args']['url'].startswith('redis'), \
            'Must have a redis url set to use a redis backend'

    # make the region
    reg = make_region(key_mangler=key_mangler).configure(
        cache_backend['name'],
        expiration_time=expiration,
        arguments=cache_backend['args']
    )
    return reg


# make a default cache region
regions['default'] = make_new_region()

# make a null cache for tests and dumps
regions['null'] = make_new_region(backend='null')

# make cache regions for NSA tables
for mpl in config._allowed_releases.keys():
    nsacache = 'nsa_{0}'.format(mpl.lower().replace('-', ''))
    regions[nsacache] = make_new_region(name=nsacache)

backend = 'file' if not redis or not os.environ.get('SESSION_REDIS') else 'redis'

# make a maps redis cache
regions['maps'] = make_new_region(backend=backend)

# make a modelcube redis cache
regions['models'] = make_new_region(backend=backend)
